from norminette.rules import Rule

assigns = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "ASSIGN",
]


class CheckVariableDeclaration(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsVarDeclaration"]

    def run(self, context):
        """
        Variables can be declared as global or in the scope of a function
        Only static variables, global variables, and constants can be initialised at declaration.
        Each variable must be declared on a separate line
        """
        i = 0
        static_or_const = False
        passed_assign = False
        if context.scope.name == "Function":
            context.scope.vars += 1
            if context.scope.vars > 5:
                context.new_error("TOO_MANY_VARS_FUNC", context.peek_token(i))
            if context.history[-2] != "IsBlockStart" and context.history[-2] != "IsVarDeclaration":
                context.new_error("VAR_DECL_START_FUNC", context.peek_token(i))
            elif context.scope.vdeclarations_allowed == False:
                context.new_error("VAR_DECL_START_FUNC", context.peek_token(i))
            elif context.scope.vdeclarations_allowed == None:
                context.scope.vdeclarations_allowed = True
        elif context.scope.name == "GlobalScope" or context.scope.name == "UserDefinedType":
            pass
        else:
            context.new_error("WRONG_SCOPE_VAR", context.peek_token(i))
        tmp = 0
        ret, tmp = context.check_type_specifier(tmp)
        tmp = context.skip_ws(tmp)
        tmp -= 1
        identifier = False
        while context.check_token(tmp, ["SEMI_COLON"] + assigns) is False:
            if context.check_token(tmp, "IDENTIFIER"):
                identifier = True
            tmp += 1
        if identifier == False:
            context.new_error("IMPLICIT_VAR_TYPE", context.peek_token(0))
            return False
        while context.peek_token(i) and context.check_token(i, "SEMI_COLON") is False:
            if context.check_token(i, "LPARENTHESIS") is True:
                i = context.skip_nest(i)
            if context.check_token(i, "ASSIGN") is True:
                passed_assign = True
            if context.check_token(i, ["STATIC", "CONST"]) is True:
                static_or_const = True
            if context.check_token(i, assigns) is True and static_or_const == False:
                if context.scope.name == "GlobalScope":
                    i += 1
                    continue
                context.new_error("DECL_ASSIGN_LINE", context.peek_token(i))
            if context.check_token(i, "COMMA") is True and passed_assign == False:
                context.new_error("MULT_DECL_LINE", context.peek_token(i))
            i += 1
        return False, 0



import glob
import os
import sys

import argparse
from norminette.lexer import Lexer, TokenError
from norminette.exceptions import CParsingError
from norminette.registry import Registry
from norminette.context import Context
from norminette.tools.colors import colors
from norminette import __version__

import _thread
from threading import Event
import time


has_err = False


def timeout(e, timeval=5):
    time.sleep(timeval)
    if e.is_set():
        return
    _thread.interrupt_main()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file",
        help="File(s) or folder(s) you wanna run the parser on. If no file provided, runs on current folder.",
        default=[],
        action="append",
        nargs="*",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="count",
        help="Debug output (multiple values available)",
        default=0,
    )
    parser.add_argument(
        "-o",
        "--only-filename",
        action="store_true",
        help="By default norminette displays the full path to the file, this allows to show only filename",
        default=False,
    )
    parser.add_argument("-v", "--version", action="version", version="norminette " + str(__version__))
    parser.add_argument(
        "--cfile",
        action="store",
        help="Store C file content directly instead of filename",
    )
    parser.add_argument(
        "--hfile",
        action="store",
        help="Store header file content directly instead of filename",
    )
    parser.add_argument(
        "--filename",
        action="store",
        help="Stores filename if --cfile or --hfile is passed",
    )
    parser.add_argument("-R", nargs=1, help="compatibility for norminette 2")
    args = parser.parse_args()
    registry = Registry()
    targets = []
    has_err = None
    content = None

    debug = args.debug
    if args.cfile != None or args.hfile != None:
        if args.filename:
            targets = [ args.filename ]
        else:
            targets = [ "file.c" ] if args.cfile else ["file.h"]
        content = args.cfile if args.cfile else args.hfile
    else:
        args.file = args.file[0]
        if args.file == [[]] or args.file == []:
            targets = glob.glob("**/*.[ch]", recursive=True)
        else:
            for arg in args.file:
                if os.path.exists(arg) is False:
                    print(f"'{arg}' no such file or directory")
                elif os.path.isdir(arg):
                    if arg[-1] != "/":
                        arg = arg + "/"
                    targets.extend(glob.glob(arg + "**/*.[ch]", recursive=True))
                elif os.path.isfile(arg):
                    targets.append(arg)
    event = []
    for target in targets:
        if target[-2:] not in [".c", ".h"]:
            print(f"Error: {target} is not valid C or C header file")
        else:
            try:
                event.append(Event())
                if content == None:
                    with open(target) as f:
                        try:
                            source = f.read()
                        except Exception as e:
                            print ("Error: File could not be read: ", e)
                            sys.exit(0)
                else:
                    source = content
                try:
                    lexer = Lexer(source)
                    tokens = lexer.get_tokens()
                except KeyError as e:
                    print ("Error while parsing file:", e)
                    sys.exit(0)
                if args.only_filename == True:
                    # target = target.split("/")[-1]
                    target = Path(target).name
                context = Context(target, tokens, debug, args.R)
                registry.run(context, source)
                event[-1].set()
                if context.errors:
                    has_err = True
            except TokenError as e:
                has_err = True
                print(target + f": Error!\n\t{colors(e.msg, 'red')}")
                event[-1].set()
            except CParsingError as e:
                has_err = True
                print(target + f": Error!\n\t{colors(e.msg, 'red')}")
                event[-1].set()
            except KeyboardInterrupt as e:
                event[-1].set()
                sys.exit(1)
    sys.exit(1 if has_err else 0)


if __name__ == "__main__":
    main()



from norminette.rules import Rule
import pdb
operators = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "LESS_OR_EQUAL",
    "GREATER_OR_EQUAL",
    "EQUALS",
    "NOT_EQUAL",
    "ASSIGN",
    # "COLON",
    # "SEMI_COLON",
    # "COMMA",
    # "DOT",
    "NOT",
    "MINUS",
    "PLUS",
    "MULT",
    "DIV",
    "MODULO",
    "LESS_THAN",
    "MORE_THAN",
    "ELLIPSIS",
    "INC",
    "DEC",
    "PTR",
    "AND",
    "OR",
    "BWISE_XOR",
    "BWISE_OR",
    "BWISE_NOT",
    "BWISE_AND",
    "RIGHT_SHIFT",
    "LEFT_SHIFT",
    "TERN_CONDITION",
]

assign_operators = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "ASSIGN",
]

gps_operators = []

ps_operators = [
    # operators that should be prefixed and suffixed by a space
    "RIGHT_ASSIGN",  # >>=
    "LEFT_ASSIGN",  # <<=
    "ADD_ASSIGN",  # +=
    "SUB_ASSIGN",  # -=
    "MUL_ASSIGN",  # *=
    "DIV_ASSIGN",  # /=
    "MOD_ASSIGN",  # %=
    "AND_ASSIGN",  # &=
    "XOR_ASSIGN",  # ^=
    "OR_ASSIGN",  # |=
    "LESS_OR_EQUAL",  # <=
    "GREATER_OR_EQUAL",  # >=
    "EQUALS",  # ==
    "NOT_EQUAL",  # !=
    "ASSIGN",  # =
    "DIV",  # /
    "MULT",  # *
    "MODULO",  # %
    "LESS_THAN",  # <
    "MORE_THAN",  # >
    "AND",  # &
    "OR",  # |
    "RIGHT_SHIFT",  # >>
    "LEFT_SHIFT",  # <<
    "TERN_CONDITION",  # ?
]

p_operators = [
    # operators that should only be prefixed by a space
    "ELLIPSIS"  # ...
]

s_operators = [
    # operators that should only be suffixed by a space
    "COMMA",  # ,
    #Where do i put that shit
    #"COLON",  # :

]

son_operators = [
    # operators that should only be suffixed by a space or newline
    "SEMI_COLON"  # ;
]

c_operators = [
    # operators that could be "glued" with another token ("x + *y", "5 + -5")
    "PLUS",
    "MINUS",
    "BWISE_XOR",  # ^
    "BWISE_OR",  # |
    "BWISE_AND",  # &
    "BWISE_NOT",  # ~
]

glued_operators = [
    "MULT",
    "PLUS",
    "MINUS",
    "DIV",
    "NOT",
    "BWISE_NOT"
]

spec_operators = [
    "NOT",
    "BWISE_NOT", 
    "DIV",
]

rnests = ["RPARENTHESIS", "RBRACE", "RBRACKET"]

lnests = [
    "LBRACE",
    "LBRACKET",
    "LPARENTHESIS",
]

left_auth = []

right_auth = []

whitespaces = ["NEWLINE", "SPACE", "TAB"]


class CheckOperatorsSpacing(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = [
            "IsFuncDeclaration",
            "IsFuncPrototype",
            "IsExpressionStatement",
            "IsAssignation",
            "IsControlStatement",
            "IsVarDeclaration",
            "IsFunctionCall",
            "IsDeclaration",
        ]
        self.last_seen_tkn = None

    def check_prefix(self, context, pos):
        tmp = -1

        if pos > 0 and context.check_token(pos, ["TAB", "SPACE"]):
            context.new_error("", context.peek_token(pos))
        if pos + 1 < len(context.tokens[: context.tkn_scope]) and context.peek_token(pos + 1).type == "SPACE":
            context.new_error("NO_SPC_AFR_OPR", context.peek_token(pos))

    def check_lnest(self, context, pos):
        if context.history[-1] == "IsFuncDeclaration" or context.history[-1] == "IsFuncPrototype":
            return False
        tmp = pos + 1
        # Here is `(_`
        while context.peek_token(tmp) and context.check_token(tmp, ["SPACE", "TAB"]) is True:
            tmp += 1
        if context.check_token(tmp, "NEWLINE") is False:
            if context.check_token(tmp, lnests + rnests + ["SEMI_COLON", "PTR", "DOT"]) is True and tmp != pos + 1:
                context.new_error("SPC_AFTER_PAR", context.peek_token(pos))
            elif context.check_token(tmp, lnests + rnests + ["SEMI_COLON", "PTR", "DOT"]) is False and tmp != pos + 1:
                context.new_error("NO_SPC_AFR_PAR", context.peek_token(pos))
        tmp = pos - 1
        # Here is `_(`
        while tmp >= 0 and context.check_token(tmp, ["SPACE", "TAB"]) is True:
            tmp -= 1
        if context.check_token(tmp, "NEWLINE") is False:
            if (
                context.check_token(
                    tmp,
                    lnests
                    + rnests
                    + [
                        "SEMI_COLON",
                        "PTR",
                        "DOT",
                        "INC",
                        "DEC",
                        "MULT",
                        "BWISE_AND",
                        "IDENTIFIER",
                        "SIZEOF",
                    ],
                )
                is True
                and tmp != pos - 1
            ):
                if context.check_token(tmp, ["MULT", "BWISE_AND"]) is True and context.is_operator == False:
                    context.new_error("NO_SPC_BFR_PAR", context.peek_token(pos))
            elif (
                context.check_token(
                    tmp,
                    lnests
                    + rnests
                    + [
                        "SEMI_COLON",
                        "PTR",
                        "DOT",
                        "INC",
                        "DEC",
                        "MULT",
                        "BWISE_AND",
                        "BWISE_OR",
                        "BWISE_XOR",
                        "BWISE_NOT",
                        "IDENTIFIER",
                        "SIZEOF",
                        "NOT",
                        "MINUS",
                        "PLUS",
                        "CONSTANT",
                        "CHAR_CONSTANT",
                        "STRING"
                    ],
                )
                is False
                and tmp == pos - 1
            ):
                context.new_error("SPC_BFR_PAR", context.peek_token(pos))
        return False

    def check_rnest(self, context, pos):
        if context.history[-1] == "IsFuncDeclaration" or context.history[-1] == "IsFuncPrototype":
            return False
        tmp = pos + 1
        # Here is `)_`
        while context.peek_token(tmp) and context.check_token(tmp, ["SPACE", "TAB"]) is True:
            tmp += 1
        if context.check_token(tmp, "NEWLINE") is False:
            if (
                context.check_token(tmp, lnests + rnests + ["SEMI_COLON", "PTR", "DOT", "INC", "DEC"]) is True
                and tmp != pos + 1
            ):
                context.new_error("NO_SPC_AFR_PAR", context.peek_token(pos))
            elif (
                context.check_token(
                    tmp,
                    lnests
                    + rnests
                    + [
                        "SEMI_COLON",
                        "PTR",
                        "DOT",
                        "INC",
                        "DEC",
                        "MINUS",
                        "MULT",
                        "BWISE_AND",
                        "IDENTIFIER",
                        "COMMA",
                        "STRING",
                        "CONSTANT",
                        "PLUS"
                    ],
                )
                is False
                and tmp == pos + 1
            ):
                context.new_error("SPC_AFTER_PAR", context.peek_token(pos))
        tmp = pos - 1
        # Here is `_)`
        while tmp > 0 and context.check_token(tmp, ["SPACE", "TAB"]) is True:
            tmp -= 1
        if context.check_token(tmp, "NEWLINE") is False:
            if (
                context.check_token(
                    tmp,
                    lnests
                    + rnests
                    + [
                        "SEMI_COLON",
                        "PTR",
                        "DOT",
                        "INC",
                        "DEC",
                        "MULT",
                        "BWISE_AND",
                        "IDENTIFIER",
                        "CONSTANT",
                    ],
                )
                is True
                and tmp != pos - 1
            ):
                context.new_error("NO_SPC_BFR_PAR", context.peek_token(pos))
        return False

    def check_suffix(self, context, pos):
        if pos + 1 < len(context.tokens[: context.tkn_scope]) and not context.check_token(
            pos + 1, ["SPACE", "NEWLINE", "TAB"] + glued_operators + rnests
        ):
            context.new_error("SPC_AFTER_OPERATOR", context.peek_token(pos))
        if pos > 0 and context.peek_token(pos - 1).type == "SPACE":
            context.new_error("NO_SPC_BFR_OPR", context.peek_token(pos))

    def check_glued_prefix_and_suffix(self, context, pos):
        if pos > 0 and context.peek_token(pos - 1).type != "SPACE":
            if context.check_token(pos - 1, "TAB") is True:
                tmp = -1
                while context.check_token(pos + tmp, "TAB") is True:
                    tmp -= 1
                if context.check_token(pos + tmp, ["NEWLINE", "ESCAPED_NEWLINE"] + glued_operators) is True:
                    return False, 0
            context.new_error("SPC_BFR_OPERATOR", context.peek_token(pos))
        if (
            pos + 1 < len(context.tokens[: context.tkn_scope])
            and context.check_token(pos + 1, ["SPACE", "LPARENTHESIS", "LBRACKET", "LBRACE", "NEWLINE"] + glued_operators) is False
        ):
            context.new_error("SPC_AFTER_OPERATOR", context.peek_token(pos))

    def check_prefix_and_suffix(self, context, pos):
        if pos > 0 and context.check_token(pos - 1, ["SPACE", "LPARENTHESIS", "LBRACKET"] + glued_operators) is False:
            if context.check_token(pos - 1, "TAB") is True:
                tmp = -1
                while context.check_token(pos + tmp, "TAB") is True:
                    tmp -= 1
                if context.check_token(pos + tmp, ["NEWLINE", "ESCAPED_NEWLINE"]) is True:
                    return False, 0
            if context.check_token(pos - 1, "RPARENTHESIS") and context.parenthesis_contain(context.skip_nest_reverse(pos - 1))[0] == "cast":
                return False, 0
            context.new_error("SPC_BFR_OPERATOR", context.peek_token(pos))
        if (pos + 1 < len(context.tokens[: context.tkn_scope]) and context.check_token(pos + 1, ["SPACE", "LPARENTHESIS", "RPARENTHESIS", "LBRACKET", "RBRACKET", "NEWLINE", "COMMA"] + spec_operators) is False):
            tmp = pos - 1
            while context.check_token(tmp, ['SPACE', 'TAB']):
                tmp -= 1
            if context.check_token(tmp, "RPARENTHESIS"):
                tmp = context.skip_nest_reverse(tmp)
                if context.parenthesis_contain(tmp)[0] != "cast":
                    context.new_error("SPC_AFTER_OPERATOR", context.peek_token(pos))
            elif context.check_token(tmp, glued_operators) is False and not (context.check_token(pos, ["PLUS", "MINUS"]) and context.check_token(pos + 1, "CONSTANT")):
                context.new_error("SPC_AFTER_OPERATOR", context.peek_token(pos))

    def check_glued_operator(self, context, pos):
        glued = [
            "LPARENTHESIS",
            "LBRACKET",
            "LBRACE",
        ]
        if context.check_token(pos + 1, ["SPACE", "TAB"]) is True:
            context.new_error("SPC_AFTER_OPERATOR", context.peek_token(pos))
        pos -= 1
        if context.check_token(pos, glued + ["SPACE", "TAB"] + glued_operators) is False:
            context.new_error("SPC_BFR_OPERATOR", context.peek_token(pos))
        while pos >= 0 and context.check_token(pos, ["SPACE", "TAB"]) is True:
            pos -= 1
            if pos >= 0 and context.check_token(pos, glued) is True:
                context.new_error("NO_SPC_BFR_OPR", context.peek_token(pos))

    def check_combined_op(self, context, pos):
        lpointer = [
            "SPACE",
            "TAB",
            "LPARENTHESIS",
            "LBRACKET",
            "MULT",
            "NOT",
            "RPARENTHESIS",
            "RBRACKET",
            "RBRACE",
            "MINUS",
            "PLUS",
            "BWISE_NOT",
            "BWISE_OR",
            "BWISE_AND",
           "BWISE_XOR",
        ]
        lsign = operators + ["LBRACKET"]
        i = 0
        if context.peek_token(pos).type == "MULT":
            if context.check_token(pos - 1, lpointer) == False and (context.is_glued_operator(pos - 1) is True):# or context.check_token(pos - 1, c_operators) is False):
                context.new_error("SPC_BFR_POINTER", context.peek_token(pos))
            if context.check_token(pos + 1, ["SPACE", "TAB"]):
                context.new_error("SPC_AFTER_POINTER", context.peek_token(pos))
            i = 1
            while context.peek_token(pos + i).type in ["MULT", "LPARENTHESIS"]:
                i += 1
                if context.peek_token(pos + i).type == "SPACE":
                    context.new_error("SPC_AFTER_POINTER", context.peek_token(pos + i))
                return i

    def run(self, context):
        """
        Some operators must be followed by a space,
        some must be only followed by a space,
        and the rest must be preceded and followed by a space.
        """
        self.last_seen_tkn = None
        i = 0
        while i < len(context.tokens[: context.tkn_scope]):
            if context.check_token(i, ["MULT", "BWISE_AND"]) is True:
                if context.is_operator(i) is False:
                    self.check_combined_op(context, i)
                    i += 1
                    continue
            if context.check_token(i, c_operators) is True:
                if context.is_glued_operator(i) is True:
                    self.check_glued_operator(context, i)
                else:
                    self.check_prefix_and_suffix(context, i)
                i += 1
                continue
            elif context.check_token(i, lnests) is True:
                self.check_lnest(context, i)
            elif context.check_token(i, rnests) is True:
                self.check_rnest(context, i)
            elif context.check_token(i, ps_operators) is True:
                self.check_prefix_and_suffix(context, i)
            elif context.check_token(i, gps_operators) is True:
                self.check_glued_prefix_and_suffix(context, i)
            elif context.check_token(i, s_operators) is True:
                self.check_suffix(context, i)
            elif context.check_token(i, son_operators) is True and context.check_token(i + 1, "NEWLINE") is False:
                self.check_suffix(context, i)
            elif context.check_token(i, p_operators) is True:
                self.check_prefix(context, i)
            if context.check_token(i, whitespaces) is False:
                self.last_seen_tkn = context.peek_token(i)
            i += 1
        return False, 0



class Scope:
    """
    Main scope class
    Contain various scope informations updated as the norminette runs through the file
    """

    def __init__(self, parent=None):
        self.parent = parent
        self.name = type(self).__name__
        self.lvl = (parent.lvl + 1) if parent is not None else 0
        self.indent = (parent.indent + 1) if parent is not None else 0
        self.lines = 0
        self.instructions = 0
        # ########################################################## #
        self.vdeclarations_allowed = False
        self.vars = 0
        self.vars_name = []
        self.vars_alignment = 0
        self.func_alignment = 0
        # ########################################################## #
        self.fdeclarations_allowed = False  # False everywhere but GlobalScope
        self.multiline = False
        self.header_protection = -1
        self.tmp_scope = None
        self.include_allowed = False

    def inner(self, sub):
        return sub(self)

    def outer(self):
        """
        Return outer scope (None if called on GlobalScope)
        Adds the line of current scope to parent scope
        to calculate function length or control structure length
        """
        if self.parent is not None:
            self.parent.lines += self.lines
        # print (f"{self.name} -> {self.parent.name}")
        return self.parent

    def get_outer(self):
        """
        Allows to peek to the parent scope without adding lines to
        the parent scope
        """
        return self.parent


class GlobalScope(Scope):
    """
    GlobalScope contains every other scope
    Has no parent scope (returns None)
    """

    def __init__(self):
        super().__init__()
        self.fdeclarations_allowed = True
        self.fnames = []
        self.functions = 0
        self.func_alignment = 0
        self.include_allowed = True


class Function(Scope):
    """
    Function definition scope, anything between the opening/closing braces of
    a function
    """

    def __init__(self, parent):
        super().__init__(parent)
        self.fname_pos = 0
        self.vdeclarations_allowed = None
        self.args = []


class ControlStructure(Scope):
    """
    Control structures scopes (if/else, while, for, do-while, ...), don't
    necessarily have opening/closing braces. If they don't, they can contain
    only one instruction, if that instruction creates a new sub scope, it can
    contain as many instruction as that scope can "hold"
    """

    def __init__(self, parent, multiline=False):
        super().__init__(parent)
        self.multiline = multiline


class UserDefinedType(Scope):
    """
    User defined type scope (struct, union, enum), only variables declarations
    are allowed within this scope
    """

    def __init__(self, parent, typedef=False):
        super().__init__(parent)
        self.typedef = typedef


class UserDefinedEnum(Scope):
    """
    User defined type scope (struct, union, enum), only variables declarations
    are allowed within this scope
    """

    def __init__(self, parent, typedef=False):
        super().__init__(parent)
        self.typedef = typedef


class VariableAssignation(Scope):
    """
    This isn't an 'actual' C scope, but it'll help us parse multiple
    assignations (int foo[4] = {0, 0, 0, 0};) easier.
    - Unused
    """

    pass



from norminette.rules import PrimaryRule

condition_ops = [
    "LESS_OR_EQUAL",
    "GREATER_OR_EQUAL",
    "EQUALS",
    "NOT_EQUAL",
    "INC",
    "DEC",
    "AND",
    "OR",
    "BWISE_XOR",
    "BWISE_OR",
    "BWISE_NOT",
    "BWISE_AND",
    "RIGHT_SHIFT",
    "LEFT_SHIFT",
    "DOT",
    "NOT",
    "MINUS",
    "PLUS",
    "MULT",
    "DIV",
    "MODULO",
    "LESS_THAN",
    "MORE_THAN",
]

SEPARATORS = [
    "COMMA",
    "AND",
    "OR",
    "SEMI_COLON"
]

assign_ops = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "ASSIGN",
]

types = [
    "CHAR",
    "DOUBLE",
    "ENUM",
    "FLOAT",
    "INT",
    "UNION",
    "VOID",
    "LONG",
    "SHORT",
    "SIGNED",
    "UNSIGNED",
    "STRUCT",
    "ENUM",
    "UNION",
    "CONST",
    "REGISTER",
    "STATIC",
    "VOLATILE",
]

op = [
    "MULT",
    "LPARENTHESIS",
    "RPARENTHESIS",
    "LBRACKET",
    "RBRACKET",
    "MINUS",
    "PLUS",
    "DIV",
    "INC",
    "DEC",
    "PTR",
    "DOT",
]

ws = ["SPACE", "TAB", "NEWLINE"]


class IsFunctionCall(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.primary = True
        self.priority = 80
        self.scope = []

    def run(self, context):
        """
        Catches function calls when it's in an assignation
        """
        i = context.skip_ws(0, nl=False)
        types = []
        while context.check_token(i, "LPARENTHESIS") is True:
            start = i
            typ, i = context.parenthesis_contain(i)
            types.append(typ)
            if typ == None or typ == "pointer":
                i = context.skip_ws(i + 1)
                if context.peek_token(i) is None or context.check_token(i, "NEWLINE") is True:
                    return False, 0
                # i += 1
                if len(types) > 1:
                    i = context.skip_ws(i, nl=False)
                    if context.check_token(i, SEPARATORS) is True:
                        i += 1
                        i = context.eol(i)
                        return True, i
                    else:
                        return False, 0
            elif typ == "function" or typ == "cast" or typ == "pointer":
                i += 1
                i = context.skip_ws(i)
        while context.check_token(i, ["MULT", "BWISE_AND"]):
            i += 1
        if context.check_token(i, "IDENTIFIER") is True:
            i += 1
            i = context.skip_ws(i)
            if context.check_token(i, "LPARENTHESIS") is True:
                i = context.skip_nest(i)
                while context.peek_token(i) is not None and context.check_token(i, SEPARATORS) is False:
                    i += 1
                i += 1
                i = context.eol(i)
                return True, i
        elif len(types) > 1 and typ == "cast" and (types[-2] == "function" or types[-2] == "pointer"):
            i += 1
            i = context.eol(i)
            return True, i
        return False, 0



import importlib
import os
from glob import glob

from norminette.rules.rule import PrimaryRule
from norminette.rules.rule import Rule


path = os.path.dirname(os.path.realpath(__file__))
files = glob(path + "/check_*.py")

rules = {}
primary_rules = {}

for f in files:
    mod_name = f.split(os.path.sep)[-1].split(".")[0]
    class_name = "".join([s.capitalize() for s in mod_name.split("_")])
    module = importlib.import_module("norminette.rules." + mod_name)
    rule = getattr(module, class_name)
    rule = rule()
    rules[class_name] = rule

files = glob(path + "/is_*.py")

for f in files:
    mod_name = f.split(os.path.sep)[-1].split(".")[0]
    class_name = "".join([s.capitalize() for s in mod_name.split("_")])
    module = importlib.import_module("norminette.rules." + mod_name)
    rule = getattr(module, class_name)
    primary_rules[class_name] = rule()


primary_rules = [v for k, v in sorted(primary_rules.items(), key=lambda item: -item[1].priority)]
__all__ = ["rules", "primary_rules"]



from functools import cmp_to_key

import norminette.rules as rules
from norminette.exceptions import CParsingError


def sort_errs(a, b):
    if a.col == b.col and a.line == b.line:
        return 1 if a.errno > b.errno else -1
    return a.col - b.col if a.line == b.line else a.line - b.line


class Registry:
    global has_err

    def __init__(self):
        self.rules = rules.rules
        self.primary_rules = rules.primary_rules
        self.dependencies = {}
        for k, r in self.rules.items():
            r.register(self)
        for k, v in self.dependencies.items():
            self.dependencies[k] = sorted(self.dependencies[k], reverse=True)

    def run_rules(self, context, rule):
        if rule.name.startswith("Is"):
            ret, read = rule.run(context)
        else:
            # print (rule.name)
            ret = False
            read = 0
            rule.run(context)
        # print(context.history, context.tokens[:5], rule)
        # if rule.name.startswith("Is"):
        #     print (rule.name, ret)
        if ret is True:
            context.scope.instructions += 1
            if rule.name.startswith("Is"):
                # print ("Line", context.tokens[0].pos[0], rule.name)
                context.tkn_scope = read
                context.history.append(rule.name)
            for r in self.dependencies.get(rule.name, []):
                self.run_rules(context, self.rules[r])
            if "all" in self.dependencies:
                for r in self.dependencies["all"]:
                    self.run_rules(context, self.rules[r])
            # context.history.pop(-1)
            context.tkn_scope = 0
        return ret, read

    def run(self, context, source):
        """
        Main function for each file.
        Primary rules are determined by the prefix "Is" and
        are run by order of priority as defined in each class
        Each secondary rule is then run in arbitrary order based on their
        dependencies
        """
        unrecognized_tkns = []
        while context.tokens != []:
            context.tkn_scope = len(context.tokens)
            for rule in self.primary_rules:
                if type(context.scope) not in rule.scope and rule.scope != []:
                    continue
                ret, jump = self.run_rules(context, rule)
                if ret is True:
                    if unrecognized_tkns != []:
                        if context.debug == 0:
                            raise CParsingError(
                                f"Error: Unrecognized line {unrecognized_tkns[0].pos} while parsing line {unrecognized_tkns}"
                            )
                        print("uncaught -> ", context.filename)
                        print("uncaught -> ", unrecognized_tkns)
                        unrecognized_tkns = []
                    context.dprint(rule.name, jump)
                    context.update()
                    context.pop_tokens(jump)
                    break
            # #############################################################
            else:  # Remove these one ALL  primary rules are done
                # print("#, ", context.tokens[0])
                unrecognized_tkns.append(context.tokens[0])
                context.pop_tokens(1)  # ##################################
            # #############################################################
        if unrecognized_tkns != []:
            print(context.debug)
            if context.debug > 0:
                print("uncaught ->", unrecognized_tkns)
        if context.errors == []:
            print(context.filename + ": OK!")
        else:
            print(context.filename + ": Error!")
            context.errors = sorted(context.errors + context.warnings, key=cmp_to_key(sort_errs))
            for err in context.errors:
                print(err)
            # context.warnings = sorted(context.warnings, key=cmp_to_key(sort_errs))
            # for warn in context.warnings:
            #     print (warn)


from norminette.rules import Rule


class IsTernary(Rule):
    def __init__(self):
        super().__init__()
        self.priority = 53
        self.scope = []

    def run(self, context):
        """
        Catches ternaries and raises an error
        """
        i = 0
        while context.peek_token(i) is not None and context.check_token(i, ["SEMI_COLON", "NEWLINE"]) is False:
            if context.check_token(i, "TERN_CONDITION") is True:
                while context.peek_token(i) is not None and context.check_token(i, ["SEMI_COLON", "NEWLINE"]) is False:
                    i += 1
                i += 1
                i = context.eol(i)
                return True, i
            i += 1
        return False, 0



from norminette.context import ControlStructure
from norminette.scope import Function
from norminette.context import GlobalScope
from norminette.rules import PrimaryRule

cs_keywords = [
    "DO",
    "WHILE",
    "FOR",
    "IF",
    "ELSE",
    "SWITCH",
    "CASE",
    "DEFAULT",
    "IDENTIFIER",
]
whitespaces = ["TAB", "SPACE", "NEWLINE"]


class IsControlStatement(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 70
        self.scope = [Function, ControlStructure, GlobalScope]

    def run(self, context):
        """
        Catches control statements, including for/switch
        Includes the condition, even if over multiple lines
        """
        is_id = False
        id_instead_cs = False
        i = context.skip_ws(0, nl=False)
        if context.check_token(i, cs_keywords) is False:
            return False, 0
        if context.check_token(i, "IDENTIFIER") is True:
            is_id = True
            id_instead_cs = True
        if context.check_token(i, ["SWITCH", "CASE", "DEFAULT"]) is True:
            i += 1
            i = context.skip_ws(i, nl=False)
            if context.check_token(i, "LPARENTHESIS") is True:
                i = context.skip_nest(i)
            i = context.skip_ws(i, nl=False)
            if context.check_token(i, ["CONSTANT", "IDENTIFIER"]) is True:
                i += 1
            i = context.skip_ws(i, nl=False)
            context.sub = context.scope.inner(ControlStructure)
            context.sub.multiline = False
            if context.check_token(i, "COLON") is True:
                i += 1
                i = context.eol(i)
                return True, i
            else:
                while context.check_token(i, ["LBRACE", "NEWLINE"]) is False:
                    i += 1
                i = context.eol(i)
                return True, i
        if context.check_token(i, "ELSE") is True:
            i += 1
            while context.check_token(i, ["TAB", "SPACE"]) is True:
                i += 1
            if context.check_token(i, "SEMI_COLON") is True:
                i += 1
                i = context.eol(i)
                return True, i
            if context.check_token(i, "NEWLINE") is True:
                context.sub = context.scope.inner(ControlStructure)
                context.sub.multiline = False
                i = context.eol(i)
                return True, i
            if context.check_token(i, ["IF"]) is True:
                pass
            elif context.check_token(i, ["LBRACE", "COMMENT", "MULT_COMMENT"]) is False:
                context.sub = context.scope.inner(ControlStructure)
                context.sub.multiline = False
                i = context.eol(i)
                return True, i
            else:
                context.sub = context.scope.inner(ControlStructure)
                context.sub.multiline = False
                i = context.eol(i)
                return True, i
        i += 1
        if id_instead_cs == True:
            return False, 0
        i = context.skip_ws(i, nl=False)
        if context.check_token(i, "LPARENTHESIS") is False:
            return False, 0
        i = context.skip_nest(i)
        i += 1
        tmp = context.skip_ws(i, nl=True)
        if context.check_token(tmp, "SEMI_COLON") is True:
            if is_id == True:
                return False, 0
            tmp += 1
            tmp = context.eol(tmp)
            return True, tmp
        i = context.skip_ws(i, nl=False)
        context.sub = context.scope.inner(ControlStructure)
        context.sub.multiline = False
        i = context.eol(i)
        return True, i



from norminette.rules import Rule

kw = [
    # C reserved keywords #
    "AUTO",
    "BREAK",
    "CASE",
    "CHAR",
    "CONST",
    "CONTINUE",
    "DEFAULT",
    "DO",
    "DOUBLE",
    "ELSE",
    "ENUM",
    "EXTERN",
    "FLOAT",
    "FOR",
    "GOTO",
    "IF",
    "INT",
    "LONG",
    "REGISTER",
    "RETURN",
    "SHORT",
    "SIGNED",
    "STATIC",
    "STRUCT",
    "SWITCH",
    "TYPEDEF",
    "UNION",
    "UNSIGNED",
    "VOID",
    "VOLATILE",
    "WHILE",
]


class CheckExpressionStatement(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = [
            "IsExpressionStatement",
            "IsControlStatement",
            "IsFunctionCall",
            "IsAssignation",
            "IsCast"
        ]

    def run(self, context):
        """
        C keywords (return, break, continue...) must be followed by a space, with the
        exception of sizeof
        Return values in a function must be contained in parenthesis
        """
        i = 0
        parenthesis = False
        while context.check_token(i, ["SEMI_COLON", "NEWLINE"]) is False:
            if context.check_token(i, kw) is True:
                if context.check_token(i + 1, ["SPACE", "NEWLINE", "RPARENTHESIS", "COMMENT", "MULT_COMMENT"]) is False:
                    context.new_error("SPACE_AFTER_KW", context.peek_token(i))
            if context.check_token(i, ["MULT", "BWISE_AND"]) is True and i > 0:
                if context.check_token(i - 1, "IDENTIFIER") is True:
                    context.new_error("SPACE_AFTER_KW", context.peek_token(i - 1))
            if context.check_token(i, "RETURN") is True:
                tmp = i + 1
                tmp = context.skip_ws(tmp)
                if (
                    context.check_token(tmp, "SEMI_COLON") is False
                    and context.check_token(tmp, "LPARENTHESIS") is False
                ):
                    context.new_error("RETURN_PARENTHESIS", context.peek_token(tmp))
                    return False, 0
                elif context.check_token(tmp, "SEMI_COLON") is False:
                    tmp = context.skip_nest(tmp) + 1
                    if context.check_token(tmp, "SEMI_COLON") is False:
                        context.new_error("RETURN_PARENTHESIS", context.peek_token(tmp))
                        return False, 0
            i += 1
        return False, 0



from norminette.rules import Rule

class CheckManyInstructions(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = [
            "IsAssignation",
            "IsBlockEnd",
            "IsControlStatement",
            "IsExpressionStatement",
            "IsFuncDeclaration",
            "IsFuncPrototype",
            "IsUserDefinedType",
            "IsVarDeclaration",
            "IsFunctionCall",
        ]

    def run(self, context):
        """
        Each instruction must be separated by a newline
        """
        if context.peek_token(0).pos[1] > 1:
            context.new_error("TOO_MANY_INSTR", context.peek_token(0))
            return False, 0
        # if context.history[-1] in ["IsFuncDeclaration", "IsFuncPrototype", "IsControlStatement"]:
        #     return False, 0
        # i = 0
        # while i < context.tkn_scope:
        #     if context.check_token(i, SEPARATORS) is True:
        #         context.new_error("TOO_MANY_INSTR", context.peek_token(0))
        #         return False, 0
        #     i += 1
        return False, 0



from norminette.rules import Rule


class CheckBrace(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsBlockStart", "IsBlockEnd"]

    def run(self, context):
        """
        C files must end with an empty line
        Functions can only have 25 lines
        """
        i = 0
        i = context.skip_ws(i, nl=False)
        # if context.check_token(i, ["RBRACE", "LBRACE"]) is False and context.scope.type != "GlobalScope":
        #    context.new_error("BRACE_EMPTY_LINE")
        if context.check_token(i, ["RBRACE", "LBRACE"]) is False:
            context.new_error("EXPECTED_BRACE", context.peek_token(i))
            return False, 0
        i += 1
        i = context.skip_ws(i, nl=False)
        if context.check_token(i, "NEWLINE") is False or context.check_token(i, "NEWLINE") is None:
            if context.scope.name == "UserDefinedType" or context.scope.name == "UserDefinedEnum":
                i = context.skip_ws(i, nl=False)
                if context.check_token(i, "SEMI_COLON") is True:
                    return False, 0
                if context.check_token(i, "IDENTIFIER") is False:
                    context.new_error("BRACE_SHOULD_EOL", context.peek_token(i - 1))
                else:
                    i += 1
                    if context.check_token(i, "SEMI_COLON") is False:
                        context.new_error("BRACE_SHOULD_EOL", context.peek_token(i - 1))
            else:
                context.new_error("BRACE_SHOULD_EOL", context.peek_token(i - 1))
        if context.scope.name == "Function" and context.scope.lines > 26:
            context.new_error("TOO_MANY_LINES", context.peek_token(0))
        return False, 0



from norminette.lexer import Token
from norminette.rules import Rule

import pdb

type_specifiers = ["CHAR", "DOUBLE", "ENUM", "FLOAT", "INT", "UNION", "VOID", "SHORT"]

misc_specifiers = ["CONST", "REGISTER", "STATIC", "STRUCT", "VOLATILE"]

size_specifiers = ["LONG", "SHORT"]

sign_specifiers = ["SIGNED", "UNSIGNED"]

whitespaces = ["SPACE", "TAB", "NEWLINE"]

arg_separator = ["COMMA", "CLOSING_PARENTHESIS"]


class CheckFuncArgumentsName(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsFuncDeclaration", "IsFuncPrototype"]

    def check_arg_format(self, context, pos):
        """
        A valid argument contains either:
        - a type specifier AND a paramater name (IDENTIFIER)
        - an ellipsis
        """

        i = context.skip_ws(pos)
        p = 0
        stop = ["COMMA", "RPARENTHESIS"]
        if context.check_token(i, ["COMMENT", "MULT_COMMENT"]):
            context.new_error("WRONG_SCOPE_COMMENT", context.peek_token(i))
            i += 1
        # if context.check_token(i, "NEWLINE"):
        # context.new_error("NEWLINE_IN_DECL", context.peek_token(i))
        # i += 1
        if context.check_token(i, "ELLIPSIS"):
            i += 1
            if context.peek_token(i).type in stop:
                i += 1
            return i
        ret, i = context.check_type_specifier(i)
        has_tab = False
        while context.check_token(i, ["SPACE", "TAB"]):
            if context.check_token(i, "TAB") is True and has_tab == False:
                context.new_error("TAB_INSTEAD_SPC", context.peek_token(i))
                has_tab = True
            i += 1

        if ret == False:
            context.new_error("ARG_TYPE_UKN", context.peek_token(i))
            return -1
        while context.peek_token(i) is not None and context.check_token(i, ["LPARENTHESIS"] + whitespaces):
            if context.check_token(i, "LPARENTHESIS") is True:
                p += 1
            if context.check_token(i, "RPARENTHESIS") is True:
                p -= 1
            i += 1

        if ret is True:
            i = context.skip_misc_specifier(i)
            ret, i = context.check_identifier(i)
            if ret is False:
                context.new_error("MISSING_IDENTIFIER", context.peek_token(i - 1))
            else:
                i += 1
                i = context.skip_ws(i)
            while context.peek_token(i) is not None and i < context.arg_pos[1]:
                if context.check_token(i, stop) is True:
                    if context.check_token(i, "RPARENTHESIS") is True and p > 0:

                        p -= 1
                    else:
                        break
                if context.check_token(i, "LPARENTHESIS"):
                    i = context.skip_nest(i)
                i += 1
            i += 1

        else:
            while context.peek_token(i) is not None and context.peek_token(i).type not in stop:
                i += 1
            i += 1
        return i

    def no_arg_func(self, context, pos):
        i = context.skip_ws(pos)
        if context.check_token(i, "VOID"):
            i += 1
            i = context.skip_ws(i)
            if context.check_token(i, "RPARENTHESIS"):
                return True
        elif context.check_token(i, "RPARENTHESIS"):
            context.new_error("NO_ARGS_VOID", context.peek_token(i))
            return True
        return False

    def run(self, context):
        """
        Empty functions arguments must use void
        """
        i = context.arg_pos[0] + 1
        ret = self.no_arg_func(context, i)
        if ret is True:
            return False, 0
        while i < context.arg_pos[1]:
            i = context.skip_ws(i)
            if context.check_token(i, "NEWLINE"):
                i += 1
                continue
            if context.check_token(i, "LPARENTHESIS"):
                p = 1
                while p:
                    if context.peek_token(i) is not None:
                        if context.check_token(i, "LPARENTHESIS"):
                            p += 1
                        elif context.check_token(i, "RPARENTHESIS"):
                            p -= 1
                    else:
                        break
                    i += 1
            elif context.check_token(i, "LBRACKET") is True:
                i = context.skip_nest(i) + 1
            elif context.check_token(i, "COMMA") is True:
                i += 1
            elif context.check_token(i, "RPARENTHESIS") is not True:
                i = self.check_arg_format(context, i)
                if i == -1:
                    return False, 0
        return False, 0



from norminette.context import ControlStructure
from norminette.scope import UserDefinedEnum
from norminette.scope import UserDefinedType
from norminette.scope import VariableAssignation
from norminette.rules import PrimaryRule


class IsBlockEnd(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 54
        self.scope = []

    def check_udef_typedef(self, context, pos):
        i = context.skip_ws(pos)
        if context.check_token(i, "IDENTIFIER") is False:
            return False, 0
        while context.check_token(i, ["IDENTIFIER", "SPACE", "TAB"]):
            i += 1
        i = context.skip_ws(i)
        if context.check_token(i, "SEMI_COLON") is False:
            return False, 0
        i += 1
        return True, i

    def run(self, context):
        """
        Catches RBRACE tokens.
        Handles scope related stuff: Exiting a scope is done here and in registry.py
        Scope is calculated AFTER the rules have run for this primary rule
        """
        i = context.skip_ws(0)
        if context.check_token(i, "RBRACE") is False:
            return False, 0
        if type(context.scope) != ControlStructure:
            context.sub = context.scope.outer()
        else:
            context.scope.multiline = False
        i += 1
        if type(context.scope) in (UserDefinedType, UserDefinedEnum):
            i = context.skip_ws(i)
            if context.check_token(i, "TYPEDEF") is True:
                i += 1
                ret, i = self.check_udef_typedef(context, i)
                i = context.eol(i)
                return ret, i
            elif context.check_token(i, "SEMI_COLON") is True:
                i += 1
                i = context.eol(i)
                return True, i

            ret, i = self.check_udef_typedef(context, i)
            i = context.eol(i)
            return ret, i

        if type(context.scope) is VariableAssignation:
            i = context.skip_ws(i)
            if context.check_token(i, "SEMI_COLON"):
                # Fatal err?
                return False, 0
            i += 1
            i = context.eol(i)
            return True, i
            pass
        if type(context.scope) is ControlStructure:
            pass
        i = context.eol(i)
        return True, i



from norminette.rules import PrimaryRule


cs_keywords = ["DO", "WHILE", "FOR", "IF", "ELSE", "SWITCH"]
whitespaces = ["TAB", "SPACE", "NEWLINE"]


class IsEmptyLine(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 65
        self.scope = []

    def run(self, context):
        """
        Catches empty line
        BUG: Catches end of line token on unrecognized line
        """
        i = 0
        while context.check_token(i, ["SPACE", "TAB"]) is True:
            i += 1
        if context.check_token(i, "NEWLINE") is True:
            i = context.eol(i)
            return True, i
        return False, 0



from norminette.rules import PrimaryRule
from norminette.scope import UserDefinedType, GlobalScope, UserDefinedEnum
import pdb

utypes = ["TYPEDEF", "UNION", "STRUCT", "ENUM"]


class IsUserDefinedType(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 45
        self.scope = [GlobalScope, UserDefinedType]

    def typedef(self, context, pos):
        i = context.skip_ws(pos)
        if "TYPEDEF" not in [tkn.type for tkn in context.tokens[:i]]:
            return False, pos
        ret, i = context.check_identifier(i)
        if ret is False:
            i += 1
            if context.check_token(i, "LBRACE"):
                return True, i
            return False, pos
        i += 1
        return True, i

    def utype_definition(self, context, pos):
        utypes = ["STRUCT", "ENUM", "UNION"]
        if not [tkn for tkn in context.tokens[:pos] if tkn.type in utypes]:
            return False, pos
        return True, pos
        i = context.skip_ws(i)
        return ret, i

    def run(self, context):
        """
        Catches user type definitions
        Can include the whole type definition in case it's a structure
        Variable declarations aren't included
        """
        i = context.skip_ws(0, nl=False)
        enum = False
        p = 0
        ids = []
        while context.peek_token(i) is not None:
            if context.check_token(i, utypes) is True and p <= 0:
                break
            if context.check_token(i, "LPARENTHESIS") is True:
                p += 1
            if context.check_token(i, "RPARENTHESIS") is True:
                p -= 1
            if context.check_token(i, ["NEWLINE", "SEMI_COLON"]) is True:
                return False, 0
            i += 1
        if context.peek_token(i) is None:
            return False, 0
        p = 0
        while context.peek_token(i):
            if context.check_token(i, "LPARENTHESIS") is True:
                p += 1
            if context.check_token(i, "RPARENTHESIS") is True:
                p -= 1
            if context.check_token(i, "ENUM") is True:
                enum = True
            if context.check_token(i, ["NEWLINE", "SEMI_COLON"]) is True and p == 0:
                break
            if context.check_token(i, "IDENTIFIER"):
                ids.append(context.peek_token(i))
            i += 1
        if context.check_token(i, "NEWLINE") is True and p <= 0:
            if enum == True:
                context.sub = context.scope.inner(UserDefinedEnum)
            else:
                context.sub = context.scope.inner(UserDefinedType)
            i = context.eol(i)
            return True, i
        elif context.check_token(i, "SEMI_COLON") is True:
            i += 1
            context.scope.vars_name.append(ids[-1])
            i = context.eol(i)
            return True, i



from norminette.rules import Rule


class CheckDeclaration(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsDeclaration"]

    def run(self, context):
        """
        Checks for nl in declarations
        """
        # i = context.skip_ws(0)
        # while context.peek_token(i) and context.check_token(i, "SEMI_COLON") is False:
        # if context.check_token(i, "NEWLINE") is True:
        # context.new_error("NEWLINE_IN_DECL", context.peek_token(i))
        # i += 1
        return False, 0



from norminette.rules import Rule


class CheckFunctionsCount(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsFuncDeclaration"]

    def run(self, context):
        """
        Each file cannot contain more than 5 function
        """
        if context.scope != None and context.scope.name == "GlobalScope":
            if context.scope.functions > 5:
                context.new_error("TOO_MANY_FUNCS", context.peek_token(0))
        return False, 0



from norminette.rules import PrimaryRule

pp_keywords = [
    "PRAGMA",
    "INCLUDE",
    "UNDEF",
    "DEFINE",
    "#IF",
    "ELIF",
    "#ELSE",
    "IFDEF",
    "IFNDEF",
    "ENDIF",
    "ERROR",
    "WARNING",
    "IMPORT",
]

whitespaces = ["TAB", "SPACE", "NEWLINE"]


class IsPreprocessorStatement(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 85
        self.scope = []

    def run(self, context):
        """
        Catches any kind of preprocessor statements
        Handles indentation related informations
        """
        i = context.skip_ws(0)
        if context.check_token(i, pp_keywords) is True:
            if context.peek_token(i).value is None or context.peek_token(i).value.startswith("#") is False:
                return False, 0
            if context.check_token(i, ["IFDEF", "IFNDEF", "#IF", "#ELIF"]):
                context.preproc_scope_indent += 1
            elif context.check_token(i, "ENDIF") and context.preproc_scope_indent > 0:
                context.preproc_scope_indent -= 1
            i += 1
        else:
            return False, 0
        i = context.eol(i)
        return True, i



errors = {
    "SPC_INSTEAD_TAB": "Spaces at beginning of line",
    "TAB_INSTEAD_SPC": "Found tab when expecting space",
    "CONSECUTIVE_SPC": "Two or more consecutives spaces",
    "SPC_BFR_OPERATOR": "missing space before operator",
    "SPC_AFTER_OPERATOR": "missing space after operator",
    "NO_SPC_BFR_OPR": "extra space before operator",
    "NO_SPC_AFR_OPR": "extra space after operator",
    "SPC_AFTER_PAR": "Missing space after parenthesis (brace/bracket)",
    "SPC_BFR_PAR": "Missing space before parenthesis (brace/bracket)",
    "NO_SPC_AFR_PAR": "Extra space after parenthesis (brace/bracket)",
    "NO_SPC_BFR_PAR": "Extra space before parenthesis (brace/bracket)",
    "SPC_AFTER_POINTER": "space after pointer",
    "SPC_LINE_START": "Unexpected space/tab at line start",
    "SPC_BFR_POINTER": "bad spacing before pointer",
    "SPACE_BEFORE_FUNC": "space before function name",
    "TOO_MANY_TABS_FUNC": "extra tabs before function name",
    "TOO_MANY_TABS_TD": "extra tabs before typedef name",
    "MISSING_TAB_FUNC": "missing tab before function name",
    "MISSING_TAB_VAR": "missing tab before variable name",
    "TOO_MANY_TAB_VAR": "extra tab before variable name",
    "LINE_TOO_LONG": "line too long",
    "EXP_PARENTHESIS": "Expected parenthesis",
    "MISSING_IDENTIFIER": "missing type qualifier or identifier in function arguments",
    "FORBIDDEN_CHAR_NAME": "user defined identifiers should contain only lowercase characters, \
digits or '_'",
    "TOO_FEW_TAB": "Missing tabs for indent level",
    "TOO_MANY_TAB": "Extra tabs for indent level",
    "SPACE_REPLACE_TAB": "Found space when expecting tab",
    "TAB_REPLACE_SPACE": "Found tab when expecting space",
    "TOO_MANY_LINES": "Function has more than 25 lines",
    "SPACE_EMPTY_LINE": "Space on empty line",
    "SPC_BEFORE_NL": "Space before newline",
    "TOO_MANY_INSTR": "Too many instructions on a single line",
    "PREPROC_UKN_STATEMENT": "Unrecognized preprocessor statement",
    "PREPROC_START_LINE": "Preprocessor statement not at the beginning of the line",
    "PREPROC_CONSTANT": "Preprocessor statement must only contain constant defines",
    "PREPROC_EXPECTED_EOL": "Expected EOL after preprocessor statement",
    "PREPROC_BAD_INDENT": "Bad preprocessor indentation",
    "USER_DEFINED_TYPEDEF": "User defined typedef must start with t_",
    "STRUCT_TYPE_NAMING": "Structure name must start with s_",
    "ENUM_TYPE_NAMING": "Enum name must start with e_",
    "UNION_TYPE_NAMING": "Union name must start with u_",
    "GLOBAL_VAR_NAMING": "Global variable must start with g_",
    "GLOBAL_VAR_DETECTED": "Global variable present in file. Make sure it is a reasonable choice.",
    "EOL_OPERATOR": "Logic operator at the end of line",
    "EMPTY_LINE_FUNCTION": "Empty line in function",
    "EMPTY_LINE_FILE_START": "Empty line at start of file",
    "EMPTY_LINE_FUNCTION": "Empty line in function",
    "EMPTY_LINE_EOF": "Empty line at end of file",
    "WRONG_SCOPE_VAR": "Variable declared in incorrect scope",
    "IMPLICIT_VAR_TYPE": "Missing type in variable declaration",
    "VAR_DECL_START_FUNC": "Variable declaration not at start of function",
    "TOO_MANY_VARS_FUNC": "Too many variables declarations in a function",
    "TOO_MANY_FUNCS": "Too many functions in file",
    "BRACE_SHOULD_EOL": "Expected newline after brace",
    "CONSECUTIVE_NEWLINES": "Consecutive newlines",
    "NEWLINE_PRECEDES_FUNC": "Functions must be separated by a newline",
    "NL_AFTER_VAR_DECL": "Variable declarations must be followed by a newline",
    "NL_AFTER_PREPROC": "Preprocessor statement must be followed by a newline",
    "MULT_ASSIGN_LINE": "Multiple assignations on a single line",
    "MULT_DECL_LINE": "Multiple declarations on a single line",
    "DECL_ASSIGN_LINE": "Declaration and assignation on a single line",
    "FORBIDDEN_CS": "Forbidden control structure",
    "SPACE_AFTER_KW": "Missing space after keyword",
    "RETURN_PARENTHESIS": "Return value must be in parenthesis",
    "EXP_SEMI_COLON": "Expected semicolon",
    "EXP_TAB": "Expected tab",
    "NO_ARGS_VOID": "Empty function argument requires void",
    "MISALIGNED_VAR_DECL": "Misaligned variable declaration",
    "MISALIGNED_FUNC_DECL": "Misaligned function declaration",
    "WRONG_SCOPE_COMMENT": "Comment is invalid in this scope",
    "MACRO_NAME_CAPITAL": "Macro name must be capitalized",
    "ASSIGN_IN_CONTROL": "Assignment in control structure",
    "VLA_FORBIDDEN": "Variable length array forbidden",
    "TOO_MANY_ARGS": "Function has more than 4 arguments",
    "INCLUDE_HEADER_ONLY": ".c file includes are forbidden",
    "INCLUDE_START_FILE": "Include must be at the start of file",
    "HEADER_PROT_ALL": "Header protection must include all the instructions",
    "HEADER_PROT_NAME": "Wrong header protection name",
    "TERNARY_FBIDDEN": "Ternaries are forbidden",
    "TOO_MANY_VALS": "Too many values on define",
    "NEWLINE_IN_DECL": "Newline in declaration",
    "MULT_IN_SINGLE_INSTR": "Multiple instructions in single line control structure",
    "NEWLINE_DEFINE": "Newline in define",
    "MISSING_TYPEDEF_ID": "Missing identifier in typedef declaration",
    "LABEL_FBIDDEN": "Label statements are forbidden",
    "PREPROC_GLOBAL": "Preprocessors can only be used in the global scope",
    "WRONG_SCOPE_FCT": "Function prototype in incorrect scope",
    "WRONG_SCOPE": "Statement is in incorrect scope",
    "INCORRECT_DEFINE": "Incorrect values in define",
    "BRACE_NEWLINE": "Expected newline before brace",
    "EXP_NEWLINE": "Expected newline after control structure",
    "ARG_TYPE_UKN": "Unrecognized variable type",
    "COMMENT_ON_INSTR": "Comment must be on its own line",
    "COMMA_START_LINE": "Comma at line start",
    "MIXED_SPACE_TAB": "Mixed spaces and tabs",
    "ATTR_EOL": "Function attribute must be at the end of line",
    "INVALID_HEADER": "Missing or invalid 42 header",
}


class NormError:
    def __init__(self, errno, line, col=None):
        self.errno = errno
        self.line = line
        self.col = col
        if col is not None:
            self.error_pos = f"(line: {(str(self.line)).rjust(3)}, col: {(str(self.col)).rjust(3)}):\t"
        else:
            self.error_pos = f"(line: {(str(self.line)).rjust(3)}):\t "
        self.prefix = f"Error: {self.errno:<20} {self.error_pos:>21}"
        self.error_msg = f"{errors.get(self.errno, 'ERROR NOT FOUND')}"

    def __str__(self):
        return self.prefix + self.error_msg

class NormWarning:
    def __init__(self, errno, line, col=None):
        self.errno = errno
        self.line = line
        self.col = col
        if col is not None:
            self.error_pos = f"(line: {(str(self.line)).rjust(3)}, col: {(str(self.col)).rjust(3)}):\t"
        else:
            self.error_pos = f"(line: {(str(self.line)).rjust(3)}):\t "
        self.prefix = f"Notice: {self.errno:<20} {self.error_pos:>21}"
        self.error_msg = f"{errors.get(self.errno, 'WARNING NOT FOUND')}"

    def __str__(self):
        return self.prefix + self.error_msg



from norminette.rules import PrimaryRule


class IsComment(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 90
        self.scope = []

    def run(self, context):
        """
        Catches comments tokens
        """
        i = context.skip_ws(0)
        if context.check_token(i, ["MULT_COMMENT", "COMMENT"]) is True:
            i += 1
            i = context.eol(i)
            return True, i
        return False, 0



from norminette.context import ControlStructure
from norminette.rules import PrimaryRule
from norminette.scope import GlobalScope, UserDefinedEnum, Function, UserDefinedType, VariableAssignation


class IsBlockStart(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 55
        self.scope = [
            Function,
            UserDefinedType,
            VariableAssignation,
            ControlStructure,
            UserDefinedEnum,
            GlobalScope,
        ]

    def run(self, context):
        """
        Catches LBRACE tokens
        Creates new scope based on previous instruction or set it to multiline if it
        is a control statement
        """
        i = context.skip_ws(0, nl=False)
        if context.check_token(i, "LBRACE") is False:
            return False, 0
        i += 1
        hist = context.history
        lines = context.scope.lines
        for item in hist[::-1]:
            if item == "IsEmptyLine" or item == "IsComment" or item == "IsPreprocessorStatement":
                lines -= 1
                continue
            if (
                item
                not in [
                    "IsControlStatement",
                    "IsFuncDeclaration",
                    "IsUserDefinedType",
                ]
                or (item in ["IsControlStatement", "IsFuncDeclaration", "IsUserDefinedType"] and lines >= 1)
            ):
                context.sub = context.scope.inner(ControlStructure)
                context.sub.multiline = True
                break
            else:
                context.scope.multiline = True
            break

        tmp = i
        # while context.peek_token(tmp) and (context.check_token(tmp, ["NEWLINE"])) is False:
        #    tmp += 1
        tmp = context.eol(tmp)
        if context.peek_token(tmp) is not None:
            i = tmp
        return True, i



from norminette.rules import Rule

ALLOWED_PREPROC = [
    "DEFINE",
    "IFNDEF",
    "IFDEF",
    "#IF",
    "ELIF",
    "#ELSE",
    "ENDIF",
    "INCLUDE",
    "WARNING",
    "UNDEF",
    "ERROR"
]
TOO_MUCH_INDENT = ["IFNDEF", "IFDEF", "ELIF", "#IF", "#ELSE"]


class CheckPreprocessorIndent(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsPreprocessorStatement"]

    def get_space_number(self, val):
        val = val[1:]
        spaces = 0
        for i in val:
            if i == " ":
                spaces += 1
            else:
                return spaces

    def run(self, context):
        """
        Preprocessor statements must be indented by an additionnal space for each #ifdef/#ifndef/#if
        statement.
        Structure is `#{indentation}preproc_statement`
        Preprocessor must always be at the start of the line
        """
        i = 0
        i = context.skip_ws(i)
        tken = context.peek_token(i)
        current_indent = context.preproc_scope_indent
        if context.peek_token(i).pos[1] != 1:
            context.new_error("PREPROC_START_LINE", context.peek_token(0))
        tken = context.peek_token(i)
        if context.check_token(i, ALLOWED_PREPROC) is False:
            context.new_error("PREPROC_UKN_STATEMENT", context.peek_token(i))
        if context.check_token(i, TOO_MUCH_INDENT) is True:
            current_indent -= 1
        if current_indent < 0:
            current_indent = 0
        fmt = ""
        val = tken.value[1:] if tken.value else tken.type
        spaces = self.get_space_number(tken.value if tken.value else tken.type)
        if current_indent != spaces:
            context.new_error("PREPROC_BAD_INDENT", context.peek_token(i))

        i += 1
        tken = context.peek_token(i)
        if tken is not None and tken.type not in ["NEWLINE", "COMMENT", "MULT_COMMENT"]:
            context.new_error("PREPROC_EXPECTED_EOL", context.peek_token(i))
        return False, 0



from norminette.rules import Rule
import re

class CheckHeader(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = []

    def parse_header(self, context):
        if context.check_token(0, "MULT_COMMENT") is False:
            # print ("Missing or invalid header. Header are being reintroduced as a mandatory part of your files. This is not yet an error.")
            context.new_error("INVALID_HEADER", context.peek_token(0))
            context.header_parsed = True
            return
        context.header += context.peek_token(0).value + '\n'

    def check_header(self, context):
        #val = r"\/\* \*{74} \*\/\n\/\*.*\*\/\n\/\*.*\*\/\n\/\*.{3}([^ ]*).*\*\/\n\/\*.*\*\/\n\/\*   By: ([^ ]*).*\*\/\n\/\*.*\*\/\n\/\*   Created: ([^ ]* [^ ]*) by ([^ ]*).*\*\/\n\/\*   Updated: ([^ ]* [^ ]*) by ([^ ]*).*\*\/\n\/\*.*\*\/\n\/\* \*{74} \*\/\n"
        val_no_check_nl = r"\/\* \*{74} \*\/.\/\*.*\*\/.\/\*.*\*\/.\/\*.{3}([^ ]*).*\*\/.\/\*.*\*\/.\/\*   By: ([^ ]*).*\*\/.\/\*.*\*\/.\/\*   Created: ([^ ]* [^ ]*) by ([^ ]*).*\*\/.\/\*   Updated: ([^ ]* [^ ]*) by ([^ ]*).*\*\/.\/\*.*\*\/.\/\* \*{74} \*\/."

        #correct_header = re.match(val, context.header)
        regex = re.compile(val_no_check_nl, re.DOTALL)
        #correct_header_no_nl = re.match(val_no_check_nl, context.header)
        correct_header_no_nl = regex.search(context.header)
        if correct_header_no_nl is None:
            # print ("Missing or invalid header. Header are being reintroduced as a mandatory part of your files. This is not yet an error.")
            context.new_error("INVALID_HEADER", context.peek_token(0))
        #else:
        #    print (correct_header.group(1,2,3,4,5,6))

    def run(self, context):
        """
            Header checking. Just a warning for now. Does not trigger moulinette error
        """
        if context.header_parsed == True:
            return False, 0
        elif context.history[-1] == "IsComment" and context.header_parsed == False:
            self.parse_header(context)
            context.header_started = True
        elif context.history[-1] != "IsComment" and context.header_started == True:
            self.check_header(context)
            context.header_parsed = True
        elif context.header_started == False and context.header_parsed == False and context.history[-1] != "IsComment":
            context.new_error("INVALID_HEADER", context.peek_token(0))
            # print ("Missing or invalid header. Header are being reintroduced as a mandatory part of your files. This is not yet an error.")
            context.header_parsed = True



from norminette.rules import Rule

allowed_on_comment = ["COMMENT", "MULT_COMMENT", "SPACE", "TAB"]


class CheckComment(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = []

    def run(self, context):
        """
        Comments are only allowed in GlobalScope.
        """
        i = context.skip_ws(0)
        has_comment = False
        while context.peek_token(i) is not None and context.check_token(i, "NEWLINE") is False:
            if context.check_token(i, allowed_on_comment) is False:
                if has_comment == True:
                    context.new_error("COMMENT_ON_INSTR", context.peek_token(i))
                    return True, i
            elif context.check_token(i, ["COMMENT", "MULT_COMMENT"]) is True:
                if context.scope.name != "GlobalScope" or context.history[-1] == "IsFuncDeclaration":
                    context.new_error("WRONG_SCOPE_COMMENT", context.peek_token(i))
                has_comment = True
            i += 1
        i = context.skip_ws(0)
        return False, 0



from norminette.rules import Rule


class CheckTernary(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = []

    def run(self, context):
        """
        Ternaries are forbidden
        """
        for i in range(0, context.tkn_scope):
            if context.check_token(i, "TERN_CONDITION") is True:
                context.new_error("TERNARY_FBIDDEN", context.peek_token(i))
        return False, 0



from norminette.context import GlobalScope
from norminette.scope import UserDefinedType
from norminette.rules import PrimaryRule

import pdb

whitespaces = ["SPACE", "TAB"]
preproc = [
    "DEFINE",
    "ERROR",
    "ENDIF",
    "ELIF",
    "IFDEF",
    "IFNDEF",
    "#IF",
    "#ELSE",
    "INCLUDE",
    "PRAGMA",
    "UNDEF",
]
assigns = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "ASSIGN",
]
misc_identifier = [
    "CONST",
    "REGISTER",
    "STATIC",
    "VOLATILE",
    "EXTERN",
    "INLINE",
    "RESTRICT",
    "SIGNED",
    "UNSIGNED",
    "TYPEDEF",
    "STRUCT",
    "ENUM",
    "UNION",
]
type_identifier = [
    "CHAR",
    "DOUBLE",
    "ENUM",
    "FLOAT",
    "INT",
    "UNION",
    "VOID",
    "LONG",
    "SHORT",
]


class IsFuncPrototype(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 82
        self.scope = [GlobalScope]

    def check_args(self, context, pos):
        i = context.skip_ws(pos)
        while context.check_token(i, whitespaces + ["RPARENTHESIS"]):
            i += 1
        if context.check_token(i, "LPARENTHESIS") is False:
            return False, pos
        p = 1
        i += 1
        while p:
            if context.check_token(i, "LPARENTHESIS"):
                p += 1
            elif context.check_token(i, "RPARENTHESIS"):
                p -= 1
            elif context.peek_token(i) is None:
                return False, 0
            i += 1
        return True, i

    def check_func_identifier(self, context, pos):
        i = context.skip_ws(pos)
        pp = 0  # Pointer operator's position
        lp = 0  # Left parenthesis counter (nesting level)
        if context.check_token(i, "IDENTIFIER"):
            i += 1
            return True, i, False

        d = ["LPARENTHESIS", "MULT"] + whitespaces
        while context.check_token(i, d):
            if context.check_token(i, "MULT") and not pp:
                pp = i
            elif pp and context.check_token(i, "LPARENTHESIS"):
                lp += 1
            i += 1
        i = context.skip_misc_specifier(i)
        if context.check_token(i, "IDENTIFIER") is False:
            return False, pos, False

        i += 1
        while context.check_token(i, ["RPARENTHESIS"] + whitespaces):
            if context.check_token(i, "RPARENTHESIS"):
                lp -= 1
            i += 1
        if pp and lp < 0 and context.check_token(i, "LPARENTHESIS") is False:
            return False, pos, False
        return True, i, (True if pp else False)

    def check_func_format(self, context):
        i = context.skip_ws(0, nl=False)
        type_id = []
        misc_id = []
        arg_start = 0
        arg_end = 0
        args = False
        identifier = None
        if context.check_token(i, "NEWLINE") is True:
            return False, 0
        while context.peek_token(i):  # and context.check_token(i, "NEWLINE") is False:
            while context.check_token(i, "IDENTIFIER") is True and context.peek_token(i).value == "__attribute__":
                i += 1
                i = context.skip_ws(i)
                i = context.skip_nest(i) + 1
                i = context.skip_ws(i)
            if context.check_token(i, misc_identifier) is True:
                misc_id.append(context.peek_token(i))
            elif context.check_token(i, type_identifier) is True:
                type_id.append(context.peek_token(i))
            if context.check_token(i, assigns + ["TYPEDEF", "COMMA", "LBRACE", "RBRACE"] + preproc) is True:
                return False, 0
            if context.check_token(i, "SEMI_COLON") is True:
                break
            elif context.check_token(i, "IDENTIFIER") is True:
                if identifier is not None:
                    type_id.append(identifier[0])
                identifier = (context.peek_token(i), i)
            if context.check_token(i, "LPARENTHESIS") is True:
                par = context.parenthesis_contain(i)
                if par[0] == "function":
                    if identifier is not None:
                        type_id.append(identifier[0])
                    while context.check_token(i, ["LPARENTHESIS", "MULT", "BWISE_AND"]):
                        i += 1
                    identifier = (context.peek_token(i), i)
                    i = context.skip_nest(i)
                elif par[0] == "pointer":
                    if identifier is not None:
                        type_id.append(identifier[0])
                    while context.check_token(i, ["LPARENTHESIS", "MULT", "BWISE_AND"]):
                        i += 1
                    identifier = (context.peek_token(i), i)
                    nxt = par[1] + 1
                    if context.check_token(nxt, ["LPARENTHESIS"]) is False:
                        return False, 0
                    i = context.skip_nest(i)
                else:
                    args = True
                    arg_start = i
                    i = context.skip_nest(i)
                    arg_end = i
                    break
            else:
                i += 1
        if len(type_id) > 0 and args == True and identifier != None:
            i = identifier[1]
            while context.check_token(i, ["LPARENTHESIS", "MULT", "BWISE_AND"]) is True:
                i += 1
            sc = context.scope
            while type(sc) != GlobalScope:
                sc = sc.outer()
            sc.fnames.append(context.peek_token(i).value)
            if context.func_alignment == 0:
                tmp = i
                while context.check_token(tmp - 1, ["LPARENTHESIS", "MULT", "BWISE_AND"]):
                    tmp -= 1
                context.func_alignment = int(context.peek_token(tmp).pos[1] / 4)
            context.fname_pos = i
            context.arg_pos = [arg_start, arg_end]
            i = arg_end
            while context.check_token(i, ["RPARENTHESIS"]) is True:
                i += 1
            i = context.skip_nest(i)
            while context.check_token(i, ["RPARENTHESIS"]) is True:
                i += 1
            i = context.skip_ws(i, nl=True)
            return True, i
        return False, 0

    def run(self, context):
        """
        Catches function prototypes
        Allows newline inside it
        End condition is SEMI_COLON token, otherwise line will be considered as
        function declaration
        """
        if type(context.scope) is not GlobalScope and type(context.scope) is not UserDefinedType:
            return False, 0
        ret, read = self.check_func_format(context)
        if ret is False:
            return False, 0
        if context.check_token(read, "IDENTIFIER") is True:
            while context.peek_token(read).value == "__attribute__":
                read += 1
                read = context.skip_ws(read)
                read = context.skip_nest(read) + 1
                read = context.skip_ws(read, nl=True)
        while context.check_token(read, ["COMMENT", "MULT_COMMENT"]) is True:
            read += 1
        read = context.skip_ws(read, nl=False)
        if context.check_token(read, "NEWLINE"):
            return False, 0

        elif context.check_token(read, "SEMI_COLON"):
            read += 1
            read = context.eol(read)
            return True, read

        return False, 0



from norminette.rules import Rule


class CheckLabel(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = []

    def run(self, context):
        """
        Goto and labels are forbidden
        """
        i = 0
        if context.scope.name != "Function":
            return False, 0
        while i < context.tkn_scope:
            if context.check_token(i, ["GOTO", "COLON"]):
                context.new_error("LABEL_FBIDDEN", context.peek_token(0))
                return False, 0
            i += 1
        return False, 0



def colors(text, *argv):
    options = {
        "bold": 1,
        "dim": 2,
        "underline": 4,
        "blink": 5,
        "reverse": 7,
        "hidden": 8,
        "reset_bold": 21,
        "reset_dim": 22,
        "reset_underlined": 24,
        "reset_blink": 25,
        "reset_reverse": 27,
        "reset_hidden": 28,
        "default_fg": 39,
        "black": 30,
        "red": 31,
        "green": 32,
        "yellow": 33,
        "blue": 34,
        "magenta": 35,
        "cyan": 36,
        "light_gray": 37,
        "dark_gray": 90,
        "light_red": 91,
        "light_green": 92,
        "light_yellow": 93,
        "light_blue": 94,
        "light_magenta": 95,
        "light_cyan": 96,
        "white": 97,
        "reset_all": 0,
    }
    reset = "\u001b[0m"
    tmp = []
    for arg in argv:
        tmp.append(str(options.get(arg, 0)))
    sep = ";"
    res = f"\u001b[{sep.join(tmp)}m{text}{reset}"
    return res



from norminette.rules import Rule

import pdb

class CheckSpacing(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = []

    def run(self, context):
        """
        Indentation (except for preprocessors) must be done with tabs
        There cannot be trailing spaces or tabs at the end of line
        """
        i = 0
        if context.history[-1] == "IsEmptyLine":
            return False, 0
        space_tab_error = False
        space_error = False
        while i in range(len(context.tokens[: context.tkn_scope])):
            if context.check_token(i, "SPACE"):
                if context.check_token(i - 1 if i > 0 else 0, "TAB"):
                    if space_tab_error == False:
                        context.new_error("MIXED_SPACE_TAB", context.peek_token(i - 1))
                        space_tab_error = True
                if context.peek_token(i).pos[1] == 1:
                    while i < context.tkn_scope and context.check_token(i, "SPACE"):
                        i += 1
                    if context.check_token(i + 1, "NEWLINE"):
                        context.new_error("SPACE_EMPTY_LINE", context.peek_token(i))
                        i += 1
                        continue
                    context.new_error("SPACE_REPLACE_TAB", context.peek_token(i))
                    continue
                i += 1
                if context.check_token(i, "SPACE"):
                    if space_error == False:
                        context.new_error("CONSECUTIVE_SPC", context.peek_token(i - 1))
                        space_error = True
                    while i < context.tkn_scope and context.check_token(i, "SPACE"):
                        i += 1
                if context.check_token(i, "NEWLINE"):
                    context.new_error("SPC_BEFORE_NL", context.peek_token(i - 1))
                if context.check_token(i, "TAB"):
                    if space_tab_error == False:
                        context.new_error("MIXED_SPACE_TAB", context.peek_token(i - 1))
                        space_tab_error = True
            elif context.check_token(i, "TAB"):
                if context.peek_token(i).pos[1] == 1:
                    while context.check_token(i, "TAB"):
                        i += 1
                    if context.check_token(i, "NEWLINE"):
                        context.new_error("SPC_BEFORE_NL", context.peek_token(i - 1))
                else:
                    i += 1
            else:
                i += 1
        return False, 0



""" Dictionary that correlates lexeme with token """

keywords = {
    # C reserved keywords #
    "auto": "AUTO",
    "break": "BREAK",
    "case": "CASE",
    "char": "CHAR",
    "const": "CONST",
    "continue": "CONTINUE",
    "default": "DEFAULT",
    "do": "DO",
    "double": "DOUBLE",
    "else": "ELSE",
    "enum": "ENUM",
    "extern": "EXTERN",
    "float": "FLOAT",
    "for": "FOR",
    "goto": "GOTO",
    "if": "IF",
    "int": "INT",
    "long": "LONG",
    "register": "REGISTER",
    "return": "RETURN",
    "short": "SHORT",
    "signed": "SIGNED",
    "sizeof": "SIZEOF",
    "static": "STATIC",
    "struct": "STRUCT",
    "switch": "SWITCH",
    "typedef": "TYPEDEF",
    "union": "UNION",
    "unsigned": "UNSIGNED",
    "void": "VOID",
    "volatile": "VOLATILE",
    "while": "WHILE",
    "inline": "INLINE",
    "NULL": "NULL",
    "restrict": "RESTRICT",
}

preproc_keywords = {
    "define": "DEFINE",
    "error": "ERROR",
    "endif": "ENDIF",
    "elif": "ELIF",
    "ifdef": "IFDEF",
    "ifndef": "IFNDEF",
    "if": "#IF",
    "else": "#ELSE",
    "include": "INCLUDE",
    "pragma": "PRAGMA",
    "undef": "UNDEF",
    "warning": "WARNING",
    "import": "IMPORT",
}

"""
__FILE__ ?add?
__LINE__ ?add?
__DATE__ ?add?
__TIME__ ?add?
__TIMESTAMP__ ?add?
"""

operators = {
    ">>=": "RIGHT_ASSIGN",
    "<<=": "LEFT_ASSIGN",
    "+=": "ADD_ASSIGN",
    "-=": "SUB_ASSIGN",
    "*=": "MUL_ASSIGN",
    "/=": "DIV_ASSIGN",
    "%=": "MOD_ASSIGN",
    "&=": "AND_ASSIGN",
    "^=": "XOR_ASSIGN",
    "|=": "OR_ASSIGN",
    "<=": "LESS_OR_EQUAL",
    ">=": "GREATER_OR_EQUAL",
    "==": "EQUALS",
    "!=": "NOT_EQUAL",
    "=": "ASSIGN",
    ";": "SEMI_COLON",
    ":": "COLON",
    ",": "COMMA",
    ".": "DOT",
    "!": "NOT",
    "-": "MINUS",
    "+": "PLUS",
    "*": "MULT",
    "/": "DIV",
    "%": "MODULO",
    "<": "LESS_THAN",
    ">": "MORE_THAN",
    "...": "ELLIPSIS",
    "++": "INC",
    "--": "DEC",
    "->": "PTR",
    "&&": "AND",
    "||": "OR",
    "^": "BWISE_XOR",
    "|": "BWISE_OR",
    "~": "BWISE_NOT",
    "&": "BWISE_AND",
    ">>": "RIGHT_SHIFT",
    "<<": "LEFT_SHIFT",
    "?": "TERN_CONDITION",
}

brackets = {
    "{": "LBRACE",
    "}": "RBRACE",
    "(": "LPARENTHESIS",
    ")": "RPARENTHESIS",
    "[": "LBRACKET",
    "]": "RBRACKET",
}

__all__ = ["brackets", "operators", "preproc_keywords", "keywords"]



from norminette.rules import PrimaryRule
import pdb

assign_ops = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "ASSIGN",
    "INC",
    "DEC",
]

SEPARATORS = [
    "COMMA",
    "AND",
    "OR",
    "SEMI_COLON"
]

types = [
    "CHAR",
    "DOUBLE",
    "ENUM",
    "FLOAT",
    "INT",
    "UNION",
    "VOID",
    "LONG",
    "SHORT",
    "SIGNED",
    "UNSIGNED",
    "STRUCT",
    "ENUM",
    "UNION",
    "CONST",
    "REGISTER",
    "STATIC",
    "VOLATILE",
]

op = [
    "MULT",
    "LPARENTHESIS",
    "RPARENTHESIS",
    "LBRACKET",
    "RBRACKET",
    "MINUS",
    "PLUS",
    "DIV",
    "PTR",
    "DOT",
]

ws = ["SPACE", "TAB", "NEWLINE"]


class IsAssignation(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.primary = True
        self.priority = 20
        self.scope = []

    def skip_ptr(self, context, pos):
        i = context.skip_ws(pos)
        while context.check_token(i, operators + ws + ["IDENTIFIER"]) is True:
            i += 1
        return i

    def check_identifier(self, context, pos):
        i = pos
        while context.check_token(i, types + ws + op + ["IDENTIFIER", "CONSTANT", "INC", "DEC"]):
            if context.check_token(i, "LBRACKET"):
                i = context.skip_nest(i)
            i += 1
        if "IDENTIFIER" in [t.type for t in context.tokens[: i + 1]]:
            return True, i
        else:
            return False, 0

    def parse_assign_right_side(self, context, i):
        while context.check_token(i, SEPARATORS) is False:
            if context.check_token(i, ["LBRACE", "LPARENTHESIS", "LBRACKET"]):
                i = context.skip_nest(i)
            i += 1
        return i


    def run(self, context):
        """
        Catches all assignation instructions
        Requires assign token
        """
        ret, i = self.check_identifier(context, 0)
        if ret is False:
            return False, 0
        tmp = 0
        while context.check_token(tmp, assign_ops) is False and tmp <= i:
            tmp += 1
        if context.check_token(tmp, assign_ops) is False:
            return False, 0
        i = tmp + 1
        #i += 1
        i = context.skip_ws(i)
        # if context.check_token(i, "LBRACE") is True:
        # i += 1
        # context.sub = context.scope.inner(VariableAssignation)
        # return True, i
        if context.scope.name == "UserDefinedEnum":
            while context.peek_token(i) and (context.check_token(i, ["COMMA", "SEMI_COLON", "NEWLINE"])) is False:
                i += 1
            i = context.eol(i)
            return True, i
        i = self.parse_assign_right_side(context, i)
        # while context.check_token(i, SEPARATORS) is False:
        #     i += 1
        #     if context.peek_token(i) is None:
        #         return False, 0
        i += 1
        i = context.eol(i)
        return True, i



import pdb

from norminette.rules import Rule
from norminette.exceptions import CParsingError


operators = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "LESS_OR_EQUAL",
    "GREATER_OR_EQUAL",
    "EQUALS",
    "NOT_EQUAL",
    "ASSIGN",
    "SEMI_COLON",
    "DOT",
    "NOT",
    "MINUS",
    "PLUS",
    "MULT",
    "DIV",
    "MODULO",
    "LESS_THAN",
    "MORE_THAN",
    "PTR",
    "AND",
    "OR",
    "BWISE_XOR",
    "BWISE_OR",
    "BWISE_NOT",
    "BWISE_AND",
    "RIGHT_SHIFT",
    "LEFT_SHIFT",
]

nest_kw = ["RPARENTHESIS", "LPARENTHESIS", "NEWLINE"]


class CheckAssignationIndent(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsAssignation", "IsFuncPrototype", "IsFunctionCall", "IsVarDeclaration"]

    def run(self, context):
        """
        Declared variables must be aligned using tabs with other variables on the same scope
        """
        i = 0
        expected = context.scope.indent
        if context.history[-1] in ["IsAssignation", "IsVarDeclaration"]:
            nest = expected + 1
        elif context.history[-1] == "IsFuncPrototype":
            nest = context.func_alignment
        else:
            nest = expected
        while context.check_token(i, ["SEMI_COLON"]) is False:
            if context.check_token(i, "NEWLINE") is True:
                if context.check_token(i - 1, operators) is True:
                    context.new_error("EOL_OPERATOR", context.peek_token(i))
                tmp = context.skip_ws(i + 1)
                if context.check_token(tmp, "COMMA"):
                    context.new_error("COMMA_START_LINE", context.peek_token(i))
                got = 0
                i += 1
                while context.check_token(i + got, "TAB") is True:
                    got += 1
                if context.peek_token(i + got) is None:
                    raise CParsingError(f"Error: Unexpected EOF l.{context.peek_token(i - 1).pos[0]}")
                if context.check_token(i + got, ["LBRACKET", "RBRACKET", "LBRACE", "RBRACE"]):
                    nest -= 1
                if got > nest or (got > nest + 1 and context.history[-1] == ["IsAssignation", "IsVarDeclaration"]):
                    context.new_error("TOO_MANY_TAB", context.peek_token(i))
                    return True, i
                elif got < nest or (got < nest - 1 and context.history[-1] in ["IsAssignation", "IsVarDeclaration"]):
                    context.new_error("TOO_FEW_TAB", context.peek_token(i))
                    return True, i
                if context.check_token(i + got, ["LBRACKET", "RBRACKET", "LBRACE", "RBRACE"]):
                    nest += 1
            if context.check_token(i, "LPARENTHESIS") is True:
                nest += 1
            if context.check_token(i, "RPARENTHESIS") is True:
                nest -= 1
            i += 1
        return False, 0



from norminette.exceptions import CParsingError
from norminette.rules import Rule
import pdb

types = [
    "STRUCT",
    "ENUM",
    "UNION",
    "INT",
    "VOID",
    "CHAR",
    "UNSIGNED",
    "CONST",
    "DOUBLE",
    "LONG",
    "SHORT",
    "STATIC",
    "IDENTIFIER",
    "SPACE",
    "TAB",
]

utypes = ["STRUCT", "ENUM", "UNION"]


class CheckUtypeDeclaration(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsUserDefinedType"]

    def run(self, context):
        """
        User defined types must respect the following rules:
            - Struct names start with s_
            - Enum names start with e_
            - Union names start with u_
            - Typedef names start with t_
        """
        i = 0
        i = context.skip_ws(i)
        tkns = context.tokens
        is_td = False
        on_newline = False
        utype = None
        contain_full_def = False
        ids = []
        while context.check_token(i, ["SEMI_COLON"]) is False and i < len(context.tokens):
            if context.check_token(i, ["SPACE", "TAB"]):
                pass
            if context.check_token(i, ["LPARENTHESIS"]) is True:
                val, tmp = context.parenthesis_contain(i)
                if val == None or val == "cast" or val == "var":
                    i = tmp
            if context.check_token(i, utypes) is True:
                utype = context.peek_token(i)
            if context.check_token(i, "TYPEDEF") is True:
                is_td = True
            if context.check_token(i, "LBRACKET") is True:
                i = context.skip_nest(i)
            if context.check_token(i, "IDENTIFIER") is True:
                if context.peek_token(i).value == "__attribute__":
                    i += 1
                    i = context.skip_ws(i)
                    i = context.skip_nest(i)
                    continue
                if context.check_token(i - 1, ["MULT", "BWISE_AND", "LPARENTHESIS"]) is True:
                    tmp = i - 1
                    while context.check_token(tmp - 1, ["MULT", "BWISE_AND", "LPARENTHESIS"]) is True and context.is_operator(tmp) == False:
                        tmp -= 1
                    # if context.check_token(tmp, "LPARENTHESIS") is True:
                    #     tmp = tmp - 1
                    #     while context.check_token(tmp, ["LPARENTHESIS"]) is True and context.is_operator(tmp) == False:
                    #         tmp -= 1
                    ids.append((context.peek_token(i), tmp))
                else:
                    ids.append((context.peek_token(i), i))
            if context.check_token(i, "LBRACE") is True:
                contain_full_def = True
                i = context.skip_nest(i)
            i += 1
        check = -1
#        print (ids, utype, contain_full_def)
        if is_td == True and len(ids) < 2 and utype != None:
            context.new_error("MISSING_TYPEDEF_ID", context.peek_token(0))
            return False, 0
        if contain_full_def == False and is_td == False and len(ids) > 1:
            check = -2
        else:
            check = -1
        if len(ids) == 0:
            return False, 0
        name = ids[0][0]
        loc = ids[check][1]
        if is_td == True:
            if ids[check][0].value.startswith("t_") is False:
                context.new_error("USER_DEFINED_TYPEDEF", context.peek_token(loc))
            if utype is not None:
                if len(ids) > 1:
                    name = ids[0][0]
                else:
                    if context.debug >= 1:
                        pass
                    elif context.debug == 0:
                        raise CParsingError(
                            f"Error: {context.filename}: Could not parse structure line {context.peek_token(0).pos[0]}"
                        )
            loc = ids[0][1]
        else:
            loc = ids[0][1]
        if is_td == False:
            if utype is not None and utype.type == "STRUCT" and name.value.startswith("s_") is False:
                context.new_error("STRUCT_TYPE_NAMING", context.peek_token(loc))
            if utype is not None and utype.type == "UNION" and name.value.startswith("u_") is False:
                context.new_error("UNION_TYPE_NAMING", context.peek_token(loc))
            if utype is not None and utype.type == "ENUM" and name.value.startswith("e_") is False:
                context.new_error("ENUM_TYPE_NAMING", context.peek_token(loc))
        if is_td or (is_td == False and contain_full_def == False):
            tmp = ids[-1][1] - 1
            tabs = 0
            while (context.check_token(tmp, "TAB")) is True and tmp > 0:
                tabs += 1
                tmp -= 1
            #if tabs > 1:
                #context.new_error("TOO_MANY_TABS_TD", context.peek_token(tmp))
            if context.check_token(tmp, "SPACE") is True:
                context.new_error("SPACE_REPLACE_TAB", context.peek_token(tmp))
            tab_error = False
            can_nl_error = False
            while tmp > 0:
                if context.check_token(tmp, "RBRACE") is True:
                    can_nl_error = True
                    tmp = context.skip_nest_reverse(tmp)
                if context.check_token(tmp, "TAB") is True and on_newline == False:
                    tab_error = True
                if context.check_token(tmp, "NEWLINE") is True and can_nl_error == False:
                    context.new_error("NEWLINE_IN_DECL", context.peek_token(ids[-1][1]))
                    can_nl_error = True
                tmp -= 1
            if tab_error:
                context.new_error("TAB_REPLACE_SPACE", context.peek_token(tmp))
        if contain_full_def == False:
            i = 0
            identifier = ids[-1][0]
            i = ids[-1][1]
            if context.check_token(i - 1, ["MULT", "BWISE_AND", "LPARENTHESIS"]) is True:
                i -= 1
                while (
                    context.check_token(i, ["MULT", "BWISE_AND", "LPARENTHESIS"]) is True
                    and context.is_operator(i) is False
                ):
                    i -= 1
            current_indent = context.peek_token(i).pos[1]
            if context.scope.vars_alignment == 0:
                context.scope.vars_alignment = current_indent
            elif context.scope.vars_alignment != current_indent:
                context.new_error("MISALIGNED_VAR_DECL", context.peek_token(0))
                return True, i
            return False, 0



import math

from norminette.rules import Rule

keywords = [
    # C reserved keywords #
    "AUTO",
    "BREAK",
    "CASE",
    "CHAR",
    "CONST",
    "CONTINUE",
    "DEFAULT",
    "DO",
    "DOUBLE",
    "ELSE",
    "ENUM",
    "EXTERN",
    "FLOAT",
    "FOR",
    "GOTO",
    "IF",
    "INT",
    "LONG",
    "REGISTER",
    "RETURN",
    "SHORT",
    "SIGNED",
    "SIZEOF",
    "STATIC",
    "STRUCT",
    "SWITCH",
    "TYPEDEF",
    "UNION",
    "UNSIGNED",
    "VOID",
    "VOLATILE",
    "WHILE",
    "IDENTIFIER",
]
eol = ["SEMI_COLON", "LPARENTHESIS"]


class CheckPrototypeIndent(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsFuncPrototype"]

    def run(self, context):
        """
        All function prototypes names must be aligned on the same indentation
        """
        i = 0
        type_identifier_nb = -1
        current_indent = 0
        id_length = 0
        buffer_len = 0
        while context.check_token(i, ["SEMI_COLON"]) is False:
            if context.check_token(i, "IDENTIFIER") is True and context.peek_token(i).value == "__attribute__":
                i += 1
                i = context.skip_ws(i)
                i = context.skip_nest(i) + 1
                type_identifier_nb += 1
            if context.check_token(i, "LPARENTHESIS") is True:
                if context.parenthesis_contain(i)[0] == "pointer":
                    i += 1
                    continue
                else:
                    break
            if context.check_token(i, keywords) is True:
                type_identifier_nb += 1
            i += 1
        i = 0
        while context.check_token(i, eol) is False:
            if context.check_token(i, "IDENTIFIER") is True and context.peek_token(i).value == "__attribute__":
                if type_identifier_nb > 0:
                    context.new_error("ATTR_EOL", context.peek_token(i))
                i += 1
                i = context.skip_ws(i)
                i = context.skip_nest(i)
                type_identifier_nb -= 1
            elif context.check_token(i, keywords) is True and type_identifier_nb > 0:
                type_identifier_nb -= 1
                if context.peek_token(i).length == 0:
                    id_length += len(str(context.peek_token(i))) - 2
                else:
                    id_length += context.peek_token(i).length
            elif context.check_token(i, "SPACE") is True and type_identifier_nb > 0:
                buffer_len += 1
            elif context.check_token(i, "SPACE") is True and type_identifier_nb == 0:
                context.new_error("SPACE_REPLACE_TAB", context.peek_token(i))
                return True, i
            elif context.check_token(i, "TAB") is True and type_identifier_nb == 0:
                if current_indent == 0:
                    current_indent = math.floor((id_length + buffer_len) / 4)
                    buffer_len = 0
                current_indent += 1
            elif context.check_token(i, "IDENTIFIER") is True and type_identifier_nb == 0:
                if context.scope.func_alignment == 0:
                    context.scope.func_alignment = current_indent
                elif current_indent != context.scope.func_alignment:
                    context.new_error("MISALIGNED_FUNC_DECL", context.peek_token(i))
                    return True, i
                return False, 0
            i += 1
        return False, 0



import math
import string

from norminette.rules import Rule
import pdb


keywords = [
    # C reserved keywords #
    "AUTO",
    "BREAK",
    "CASE",
    "CHAR",
    "CONST",
    "CONTINUE",
    "DEFAULT",
    "DO",
    "DOUBLE",
    "ELSE",
    "ENUM",
    "EXTERN",
    "FLOAT",
    "FOR",
    "GOTO",
    "IF",
    "INT",
    "LONG",
    "REGISTER",
    "RETURN",
    "SHORT",
    "SIGNED",
    "SIZEOF",
    "STATIC",
    "STRUCT",
    "SWITCH",
    "TYPEDEF",
    "UNION",
    "UNSIGNED",
    "VOID",
    "VOLATILE",
    "WHILE",
    "IDENTIFIER",
]
assigns_or_eol = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "ASSIGN",
    "SEMI_COLON",
    "NEWLINE",
    "COMMA",
]


class CheckVariableIndent(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsVarDeclaration"]

    def check_tabs(self, context):
        i = 0
        current_indent = context.scope.indent
        type_identifier_nb = -1
        has_tab = False
        line_start = True
        id_length = 0
        buffer_len = 0
        while context.check_token(i, assigns_or_eol) is False:
            if context.check_token(i, keywords) is True:
                type_identifier_nb += 1
            if context.check_token(i, ["LPARENTHESIS", "LBRACE", "LBRACKET"]) and type_identifier_nb > 0 and context.parenthesis_contain(i)[0] != "pointer":
                i = context.skip_nest(i)
            i += 1
        i = 0
        while context.check_token(i, assigns_or_eol) is False:
            if context.check_token(i, "LBRACKET") is True:
                while context.check_token(i, "RBRACKET") is False:
                    if context.check_token(i, "SIZEOF") is True:
                        i += 1
                        i = context.skip_nest(i)
                        continue
                    elif context.check_token(i, "IDENTIFIER") is True:
                        for c in context.peek_token(i).value:
                            if c in string.ascii_lowercase:
                                context.new_error("VLA_FORBIDDEN", context.peek_token(i))
                                break
                        return True, i
                    i += 1
            if context.check_token(i, keywords) is True and type_identifier_nb > 0:
                line_start = False
                type_identifier_nb -= 1
                if context.peek_token(i).length == 0:
                    id_length = len(str(context.peek_token(i))) - 2
                else:
                    id_length = context.peek_token(i).length
                current_indent += math.floor((id_length + buffer_len) / 4)
                buffer_len = 0
            elif context.check_token(i, "SPACE") is True and type_identifier_nb > 0:
                buffer_len += 1
            elif context.check_token(i, "TAB") is False and type_identifier_nb == 0:
                context.new_error("SPACE_REPLACE_TAB", context.peek_token(i))
                return True, i
            elif context.check_token(i, "TAB") is True and type_identifier_nb == 0:
                has_tab += 1
                current_indent += 1
                type_identifier_nb -= 1
            elif context.check_token(i, "TAB") and type_identifier_nb > 0 and line_start == False:
                context.new_error("TAB_REPLACE_SPACE", context.peek_token(i))
            i += 1
        return False, 0

    def run(self, context):
        """
        Each variable must be indented at the same level for its scope
        """
        i = 0
        identifier = None
        ident = [0, 0]
        ret = None
        self.check_tabs(context)
        while context.peek_token(i) and context.check_token(i, ["SEMI_COLON", "COMMA", "ASSIGN"]) is False:
            if context.check_token(i, ["LBRACKET", "LBRACE"]) is True:
                i = context.skip_nest(i)
            if context.check_token(i, "LPARENTHESIS") is True:
                ret, _ = context.parenthesis_contain(i)
            if context.check_token(i, "IDENTIFIER") is True:
                ident = (context.peek_token(i), i)
                if ret == "pointer":
                    break
            i += 1
        i = ident[1]
        identifier = ident[0]
        if context.check_token(i - 1, ["MULT", "BWISE_AND", "LPARENTHESIS"]) is True:
            i -= 1
            while (
                context.check_token(i - 1, ["MULT", "BWISE_AND", "LPARENTHESIS"]) is True
                and context.is_operator(i) is False
            ):
                i -= 1
            identifier = context.peek_token(i)
        if context.scope.vars_alignment == 0:
            context.scope.vars_alignment = identifier.pos[1]
        elif context.scope.vars_alignment != identifier.pos[1]:
            context.new_error("MISALIGNED_VAR_DECL", context.peek_token(i))
            return True, i
        return False, 0



class Rule:
    def __init__(self):
        self.name = type(self).__name__
        self.depends_on = []
        self.primary = False

    def register(self, registry):
        if self.depends_on == []:
            if "all" in registry.dependencies:
                registry.dependencies["all"].append(self.name)
            else:
                registry.dependencies["all"] = [self.name]

        for rule in self.depends_on:
            if rule in registry.dependencies:
                registry.dependencies[rule].append(self.name)
            else:
                registry.dependencies[rule] = [self.name]

class PrimaryRule(Rule):
    def __init__(self):
        super().__init__()
        self.primary = True
        self.priority = 0
        self.scope = []

    def run(self, context):
        return False, 0



from norminette.rules import Rule

types = [
    "STRUCT",
    "ENUM",
    "UNION",
]


class CheckStructNaming(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsUserDefinedType"]
        self.__i = 0

    def run(self, context):
        """
        Rewritten elsewhere
        """
        return False, 0
        self.__i += 1
        i = 0
        i = context.skip_ws(i)
        while context.check_token(i, types) is False:
            i += 1
        if context.check_token(i, "NEWLINE"):
            return False, 0
        def_type = context.peek_token(i).type
        i += 1
        i = context.skip_ws(i)
        if def_type == "STRUCT":
            if context.peek_token(i).value.startswith("s_") is False:
                context.new_error("STRUCT_TYPE_NAMING", context.peek_token(i))
        elif def_type == "ENUM":
            if context.peek_token(i).value.startswith("e_") is False:
                context.new_error("ENUM_TYPE_NAMING", context.peek_token(i))
        elif def_type == "UNION":
            if context.peek_token(i).value.startswith("u_") is False:
                context.new_error("UNION_TYPE_NAMING", context.peek_token(i))
        return False, i



from norminette.rules import PrimaryRule

cs_keywords = ["DO", "WHILE", "FOR", "IF", "ELSE", "SWITCH", "CASE", "DEFAULT"]
whitespaces = ["TAB", "SPACE", "NEWLINE"]


class IsAmbiguousDeclaration(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 0
        self.scope = []

    def run(self, context):
        """
        Catches missing semi-colon or other various missing stuff. Dev feature
        """
        i = context.skip_ws(0, nl=False)
        while context.peek_token(i) and context.check_token(i, "NEWLINE") is False:
            if context.check_token(i, ["SEMI_COLON"]) is False:
                return False, 0
            i += 1
        return True, i



from norminette.context import ControlStructure
from norminette.scope import Function
from norminette.exceptions import CParsingError
from norminette.rules import PrimaryRule
import pdb
keywords = ["BREAK", "CONTINUE", "GOTO", "RETURN"]

operators = [
    "MULT",
    "LPARENTHESIS",
    "RPARENTHESIS",
    "LBRACKET",
    "RBRACKET",
    "PTR",
    "DOT",
]

ws = ["SPACE", "TAB", "NEWLINE"]


class IsExpressionStatement(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 25
        self.scope = [Function, ControlStructure]

    def check_reserved_keywords(self, context, pos):
        if context.check_token(pos, keywords) is False:
            return False, pos
        if context.check_token(pos, "RETURN"):
            i = pos + 1
            while context.check_token(i, "SEMI_COLON") is False:
                i += 1
            i += 1
            return True, i
        elif context.check_token(pos, "GOTO"):
            i = pos + 1
            i = context.skip_ws(i)
            while context.check_token(i, ["MULT", "BWISE_AND"]) is True and context.is_operator(i) is False:
                i += 1
            if context.check_token(i, "IDENTIFIER") is False:
                if context.check_token(i, "LPARENTHESIS") is True:
                    # parse label value here
                    i = context.skip_nest(i)
                elif context.debug == 0:
                    raise CParsingError("Error: Goto statement should be followed by a label")
            i += 1
            i = context.skip_ws(i)
            i += 1
            return True, i
        else:
            i = pos + 1
            i = context.skip_ws(i)
            i += 1
            return True, i

    def check_instruction(self, context, pos):
        i = pos
        if context.check_token(i, ["IDENTIFIER"]) is False:
            return False, pos
        i += 1
        # TO DO:
        # Skip possible brackets:
        # eg: "f[0]();" is a valid instruction

        i = context.skip_ws(i)
        if context.check_token(i, "LPARENTHESIS") is False:
            return False, pos

        # Shouldn't use skip_nest, should parse manually ?
        i = context.skip_nest(i)
        i += 1
        i = context.skip_ws(i)
        if context.check_token(i, "SEMI_COLON") is False:
            return False, pos
        i += 1
        return True, i

    def check_inc_dec(self, context, pos):
        i = pos
        ret = False
        if context.check_token(i, ["INC", "DEC"]) is True:
            ret = True
            i += 1
            i = context.skip_ws(i)
        if context.check_token(i, "LPARENTHESIS") is True:
            i = context.skip_nest(i)
            i += 1
            i = context.skip_ws(i)
            if ret is False:
                if context.check_token(i, ["INC", "DEC"]) is False:
                    return False, pos
                i += 1
                i = context.skip_ws(i)
            if context.check_token(i, "SEMI_COLON") is False:
                return False, pos
            i += 1
            return True, i
        if context.check_token(i, "IDENTIFIER") is True:
            i += 1
            if ret is False:
                i = context.skip_ws(i)
                if context.check_token(i, ["INC", "DEC"]) is False:
                    return False, pos
                i += 1
                i = context.skip_ws(i)
            while context.check_token(i, "SEMI_COLON") is False:
                i += 1
            i += 1
            return True, i
        return False, pos

    def void_identifier(self, context, pos):
        if context.check_token(pos, "LPARENTHESIS") is False:
            return False, pos
        i = pos + 1
        if context.check_token(i, "VOID") is False:
            return False, pos
        i += 1
        if context.check_token(i, "RPARENTHESIS") is False:
            return False, pos
        i += 1
        i = context.skip_ws(i)
        if context.check_token(i, "IDENTIFIER") is False:
            return False, pos
        i += 1
        while context.check_token(i, "SEMI_COLON") is False:
            i += 1
        i += 1
        return True, i

    def run(self, context):
        """
        Catches expression statement by elimination
        """
        i = context.skip_ws(0)
        ret, i = self.check_instruction(context, i)
        if ret is False:
            ret, i = self.check_reserved_keywords(context, i)
            #if ret is False:
            #    ret, i = self.check_inc_dec(context, i)
            if ret is False:
                ret, i = self.void_identifier(context, i)
                if ret is False:
                    return False, 0
        i = context.eol(i)
        return True, i



from norminette.rules import Rule


forbidden_cs = ["FOR", "SWITCH", "CASE", "GOTO"]
assigns = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "ASSIGN",
]


class CheckControlStatement(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsControlStatement"]

    def check_nest(self, context, i):
        depth = 1
        i += 1
        while depth > 0:
            if context.check_token(i, "LPARENTHESIS") is True:
                depth += 1
            if context.check_token(i, "RPARENTHESIS") is True:
                depth -= 1
            if context.check_token(i, assigns) is True:
                context.new_error("ASSIGN_IN_CONTROL", context.peek_token(i))
                return -1
            if context.check_token(i, forbidden_cs) is True:
                context.new_error("FORBIDDEN_CS", context.peek_token(i))
            if context.check_token(i, "NEWLINE") is True and depth < 1:
                return
            i += 1
        return

    def run(self, context):
        """
        Forbidden control structures:
            - For
            - Switch case
            - Goto
        Assignations must be done outside of control structures
        """
        i = 0
        if context.scope.name == "GlobalScope":
            context.new_error("WRONG_SCOPE", context.peek_token(0))
        while context.check_token(i, "NEWLINE") is False:
            if context.check_token(i, "SEMI_COLON") is True:
                context.new_error("EXP_NEWLINE", context.peek_token(i))
                return True, i
            if context.check_token(i, forbidden_cs) is True:
                context.new_error("FORBIDDEN_CS", context.peek_token(i))
                return True, i
            elif context.check_token(i, "LPARENTHESIS") is True:
                if self.check_nest(context, i) == -1:
                    return True, i
            i += 1
        if i < context.tkn_scope:
            i += 1
            indent = 0
            while context.check_token(i, ["TAB"]) is True:
                i += 1
                indent += 1
            if context.check_token(i, "SEMI_COLON") is True:
                if indent > context.scope.indent + 1:
                    context.new_error("TOO_MANY_TAB", context.peek_token(i))
                if indent < context.scope.indent + 1:
                    context.new_error("TOO_FEW_TAB", context.peek_token(i))
        return False, 0



import string

from norminette.rules import Rule
from norminette.scope import GlobalScope, UserDefinedType


assigns = ["ASSIGN"]


class CheckIdentifierName(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = []

    def run(self, context):
        """
        Function can only be declared in the global scope
        User defined identifiers can only contain lowercase characters, '_' or digits
        """
        i = 0
        legal_characters = string.ascii_lowercase + string.digits + "_"
        legal_cap_characters = string.ascii_uppercase + string.digits + "_"
        if context.history[-1] == "IsFuncDeclaration":
            sc = context.scope
            if type(sc) is not GlobalScope and type(sc) is not UserDefinedType:
                context.new_error("WRONG_SCOPE_FCT", context.peek_token(0))
            while type(sc) != GlobalScope:
                sc = sc.outer()
            for c in sc.fnames[-1]:
                if c not in legal_characters:
                    context.new_error("FORBIDDEN_CHAR_NAME", context.peek_token(context.fname_pos))
        if len(context.scope.vars_name) > 0:
            for val in context.scope.vars_name[::]:
                for c in val.value:
                    if c not in legal_characters:
                        context.new_error("FORBIDDEN_CHAR_NAME", val)
                        break
                context.scope.vars_name.remove(val)
        # passed_assign = False
        # err = None
        # hist = context.history[-1]
        # while i < context.tkn_scope and context.peek_token(i) is not None:
        # if context.check_token(i, assigns) is True:
        # passed_assign = True
        #        if context.check_token(i, "IDENTIFIER") and hist in ['IsVarDeclaration']:
        # for c in context.peek_token(i).value:
        # if c not in legal_characters:
        # err = ("FORBIDDEN_CHAR_NAME", context.peek_token(i))
        # break
        # if err is not None and hist not in ['IsFuncDeclaration', 'IsFuncPrototype'] or (hist == 'IsVariable' and passed_assign == True):
        # for c in context.peek_token(i).value:
        # if c not in legal_cap_characters:
        # err = ("FORBIDDEN_CHAR_NAME", context.peek_token(i))
        # break
        # else:
        # err = None
        # if err is not None:
        # context.new_error(err[0], err[1])
        # break
        # i += 1
        return False, 0



from norminette.rules import Rule


class CheckCommentLineLen(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsComment"]

    def run(self, context):
        """
        Lines must not be over 80 characters long
        """
        i = 0
        while context.check_token(i, ["COMMENT", "MULT_COMMENT"]) is False:
            i += 1
        val = context.peek_token(i).value
        line_start = context.peek_token(0).pos[1]
        val = val.split("\n")
        for item in val:
            if len(item) + line_start > 81:
                context.new_error("LINE_TOO_LONG", context.peek_token(0))
            line_start = 0



from pathlib import Path

from norminette.lexer import Lexer
from norminette.rules import Rule
from norminette.scope import GlobalScope, Function

import pdb

class CheckPreprocessorDefine(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsPreprocessorStatement"]

    def skip_define_nest(self, i, tkns):
        eq = {
            "LPARENTHESIS": "RPARENTHESIS",
            "LBRACKET": "RBRACKET",
            "LBRACE": "RBRACE",
        }
        eq_val = eq[tkns[i].type]
        while i < len(tkns) and tkns[i].type != eq_val:
            i += 1
        return i

    def check_function_declaration(self, context):
        i = context.skip_ws(0)
        if context.check_token(i, ["#IF", "#ELSE", "IFDEF", "IFNDEF"]) is False:
            return
        context.tmp_scope = context.scope
        context.scope = context.scope.get_outer()

    def run(self, context):
        """
        Preprocessor statements must be defined only in the global scope
        Defined names must be in capital letters
        Define cannot contain newlines
        Define can only contain constant values, such as integers and strings
        """
        i = context.skip_ws(0)
        if len(context.history) > 1 and context.history[-2] == "IsFuncDeclaration":
            self.check_function_declaration(context)
        if type(context.scope) is not GlobalScope:
            if type(context.scope) == Function and context.scope.multiline == False:
                pass
            else:
                context.new_error("PREPROC_GLOBAL", context.peek_token(0))
        if context.check_token(i, "DEFINE") is True:
            val = context.peek_token(i).value.split("define", 1)[1]
        elif context.check_token(i, "IFNDEF") is True:
            val = context.peek_token(i).value.split("ifndef", 1)[1]
        else:
            return False, 0
        content = Lexer(val, context.peek_token(i).pos[0])
        tkns = content.get_tokens()
        i = 0
        identifiers = []
        # protection = context.filename.upper().split("/")[-1].replace(".", "_")
        protection = Path(context.filename).name.upper().replace(".", "_")
        for tkn in tkns:
            if tkn.type == "ESCAPED_NEWLINE":
                context.new_error("NEWLINE_DEFINE", tkn)
            elif tkn.type in ["TAB", "SPACE"]:
                if tkn.type == "TAB" and len(identifiers) == 0:
                    context.new_error("TAB_INSTEAD_SPC", tkn)
                i += 1
                continue
            elif tkn.type == "IDENTIFIER" and len(identifiers) == 0:
                if tkn.value.isupper() is False:
                    context.new_error("MACRO_NAME_CAPITAL", tkn)
                identifiers.append(tkn)
                tmp = i
                while tmp < len(tkns) - 1 and tkns[tmp].type in [
                    "SPACE",
                    "TAB",
                    "IDENTIFIER",
                ]:
                    if tkn.type == "TAB":
                        context.new_error("TAB_INSTEAD_SPC", tkn)
                    tmp += 1
                if tmp == (len(tkns) - 1) and context.filetype == "h":
                    if context.scope.header_protection == 0:
                        if identifiers[0].value == protection:
                            context.scope.header_protection = 1
                        elif identifiers[0].value != protection:
                            context.new_error("HEADER_PROT_NAME", tkns[1])
                elif (
                    context.filetype == "c"
                    and context.scope.include_allowed == True
                    and (
                        len(tkns) > tmp + 1
                        or (
                            len(tkns) == tmp + 1
                            and identifiers[0].value != protection
                            and context.scope.header_protection == -1
                        )
                    )
                ):
                    context.scope.include_allowed = False
            elif tkn.type in ["IDENTIFIER", "STRING", "CONSTANT"]:
                if context.skip_define_error == True:
                    continue
                if len(identifiers) == 1:
                    if tkn.type == "IDENTIFIER" and tkn.value.isupper() is False:
                        context.new_error("PREPROC_CONSTANT", tkn)
                    identifiers.append(tkn)
                elif len(identifiers) == 0:
                    context.new_error("INCORRECT_DEFINE", tkn)
                else:
                    context.new_error("TOO_MANY_VALS", tkn)
            elif tkn.type == "LPARENTHESIS":
                if context.skip_define_error == True:
                    continue
                if len(identifiers) == 0:
                    continue
                elif len(identifiers) == 1 and tkns[i - 1].type in ["SPACE", "TAB"]:
                    if tkn.type == "TAB":
                        context.new_error("TAB_INSTEAD_SPC", tkn)
                    continue
                else:
                    context.new_error("PREPROC_CONSTANT", tkn)
            elif tkn.type in ["LBRACKET", "LBRACE"]:
                if context.skip_define_error == True:
                    continue
                context.new_error("PREPROC_CONSTANT", tkn)
            i += 1
        if context.filetype == "h" and context.scope.header_protection != 1:
            context.new_error("HEADER_PROT_ALL_2", context.peek_token(i))
        return False, 0



from norminette.rules import Rule


operators = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "LESS_OR_EQUAL",
    "GREATER_OR_EQUAL",
    "EQUALS",
    "NOT_EQUAL",
    "ASSIGN",
    "DOT",
    "NOT",
    "MINUS",
    "PLUS",
    "MULT",
    "DIV",
    "MODULO",
    "LESS_THAN",
    "MORE_THAN",
    "PTR",
    "AND",
    "OR",
    "BWISE_XOR",
    "BWISE_OR",
    "BWISE_NOT",
    "BWISE_AND",
    "RIGHT_SHIFT",
    "LEFT_SHIFT",
]
nest_kw = ["RPARENTHESIS", "LPARENTHESIS", "NEWLINE"]


class CheckNestLineIndent(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsControlStatement", "IsExpressionStatement", "IsDeclaration"]

    def find_nest_content(self, context, nest, i):
        expected = context.scope.indent + nest
        while context.peek_token(i) is not None:
            if context.check_token(i, ["LPARENTHESIS", "LBRACE", "LBRACKET"]) is True:
                i += 1
                i = self.find_nest_content(context, nest + 1, i) + 1
            if context.check_token(i, ["RBRACE", "RBRACKET", "RPARENTHESIS"]):
                return i
            elif context.check_token(i, "NEWLINE") is True:
                if context.check_token(i - 1, operators):
                    context.new_error("EOL_OPERATOR", context.peek_token(i - 1))
                if context.check_token(i, "SEMI_COLON") is True:
                    return i
                indent = 0
                i += 1
                while context.check_token(i, "TAB") is True:
                    indent += 1
                    i += 1
                if context.check_token(i, ["RBRACE", "RBRACKET", "RPARENTHESIS"]):
                    expected -= 1
                if indent > expected:
                    context.new_error("TOO_MANY_TAB", context.peek_token(i))
                elif indent < expected:
                    context.new_error("TOO_FEW_TAB", context.peek_token(i))
                if context.check_token(i, ["RBRACE", "RBRACKET", "RPARENTHESIS"]):
                    expected += 1
            else:
                i += 1
        return i

    def run(self, context):
        """
        Each nest (parenthesis, brackets, braces) adds a tab to the general indentation
        """
        i = 0
        expected = context.scope.indent
        nest = 0
        if context.history[-1] == "IsEmptyLine":
            return False, 0
        while context.peek_token(i) and context.check_token(i, ["LPARENTHESIS", "NEWLINE"]) is False:
            i += 1
        if context.check_token(i, "NEWLINE") is True:
            return False, 0
        if context.check_token(i, "LPARENTHESIS") is True:
            nest += 1
            i += 1
            self.find_nest_content(context, nest, i)
        return False, 0



from norminette.rules import Rule


class CheckEmptyLine(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = []

    def run(self, context):
        """
        Empty line must not contains tabs or spaces
        You cannot have 2 empty lines in a row
        Your variable declarations must be followed by an empty line
        No other empty lines are allowed in functions
        You must have an empty between two functions
        """
        i = 0
        if len(context.history) == 1 and context.history[-1] == "IsEmptyLine":
            context.new_error("EMPTY_LINE_FILE_START", context.peek_token(i))
            return False, 0
        if context.scope.name != "GlobalScope" and context.history[-1] != "IsBlockStart":
            if context.history[-1] != "IsVarDeclaration" and context.scope.vdeclarations_allowed == True:
                context.scope.vdeclarations_allowed = False
                if context.history[-1] not in ["IsEmptyLine", "IsComment"]:
                    if context.history[-1] == "IsBlockEnd" and context.scope.name == "Function":
                        pass
                    else:
                        context.new_error("NL_AFTER_VAR_DECL", context.peek_token(i))
                        return True, i
        if len(context.history) > 1 and context.history[-2] == "IsPreprocessorStatement" and context.history[-1] != "IsPreprocessorStatement" and context.history[-1] != "IsEmptyLine" and context.history[-1] != "IsComment":
            context.new_error("NL_AFTER_PREPROC", context.peek_token(i))
        if context.history[-1] != "IsEmptyLine":
            return False, 0
        if context.check_token(i, "NEWLINE") is False:
            context.new_error("SPACE_EMPTY_LINE", context.peek_token(i))
        if context.history[-2] == "IsEmptyLine":
            context.new_error("CONSECUTIVE_NEWLINES", context.peek_token(i))
        if context.history[-2] != "IsVarDeclaration" and context.scope.name != "GlobalScope":
            context.new_error("EMPTY_LINE_FUNCTION", context.peek_token(i))
        if context.peek_token(i + 1) is None:
            context.new_error("EMPTY_LINE_EOF", context.peek_token(i))

        return False, 0



import string
import pdb

from norminette.lexer.dictionary import brackets
from norminette.lexer.dictionary import keywords
from norminette.lexer.dictionary import operators
from norminette.lexer.dictionary import preproc_keywords
from norminette.lexer.tokens import Token


def read_file(filename):
    with open(filename) as f:
        return f.read()


class TokenError(Exception):
    def __init__(self, pos):
        self.msg = f"Error: Unrecognized token line {pos[0]}, col {pos[1]}"

    def __repr__(self):
        return self.msg


class Lexer:
    def __init__(self, source_code, starting_line=1):
        self.src = source_code
        self.len = len(source_code)
        self.__char = self.src[0] if self.src != "" else None
        self.__pos = int(0)
        self.__line_pos = int(starting_line)
        self.__line = int(starting_line)
        self.tokens = []

    def peek_sub_string(self, size):
        return self.src[self.__pos : self.__pos + size]

    def peek_char(self):
        """Return current character being checked,
        if the character is a backslash character the following
        character is appended to the return value. It will allow us to
        parse escaped characters easier.
        """
        if self.__pos < self.len:
            if self.src[self.__pos] == "\\":
                self.__char = self.src[self.__pos : self.__pos + 2]
            else:
                self.__char = self.src[self.__pos]
        else:
            self.__char = None
        return self.__char

    def pop_char(self):
        """Pop a character that's been read by increasing self.__pos,
        for escaped characters self.__pos will be increased twice
        """
        if self.peek_char() == "\t":
            # this calculates the 'visual offset' of a tab based on it's
            # position on the line, if there's an easier way to calculate this
            # you're welcome
            self.__line_pos = int((self.__line_pos + 4 - (self.__line_pos - 1) % 4) * 5 / 5)
        else:
            self.__line_pos += 1
        if self.__pos < self.len and self.src[self.__pos] == "\\":
            self.__pos += 1
        self.__pos += 1
        return self.peek_char()

    def peek_token(self):
        return self.tokens[-1]

    def line_pos(self):
        return self.__line, self.__line_pos

    def is_string(self):
        """True if current character could start a string constant"""
        if self.peek_sub_string(2) == 'L"' or self.peek_char() == '"':
            return True
        else:
            return False

    def is_constant(self):
        """True if current character could start a numeric constant"""
        if self.peek_char() in string.digits:
            return True
        elif self.peek_char() == ".":
            for i in range(0, self.len - self.__pos):
                if self.src[self.__pos + i] == ".":
                    i += 1
                elif self.src[self.__pos + i] in "0123456789":
                    return True
                else:
                    return False
        else:
            return False

    def is_char_constant(self):
        """True if current character could start a character constant"""
        if self.peek_char() == "'" or self.peek_sub_string(2) == "L'":
            return True
        else:
            return False

    def string(self):
        """String constants can contain any characer except unescaped newlines.
        An unclosed string or unescaped newline is a fatal error and thus
        parsing will stop here.
        """
        pos = self.line_pos()
        tkn_value = ""
        if self.peek_char() == "L":
            tkn_value += self.peek_char()
            self.pop_char()
        tkn_value += self.peek_char()
        self.pop_char()
        while self.peek_char() not in [None]:
            tkn_value += self.peek_char()
            if self.peek_sub_string(2) == "\\\n":
                self.__line += 1
                self.__line_pos = 1
            if self.peek_char() == '"':
                break
            self.pop_char()
        else:
            raise TokenError(pos)
            return
        self.tokens.append(Token("STRING", pos, tkn_value))
        self.pop_char()

    def char_constant(self):
        """Char constants follow pretty much the same rule as string constants"""
        pos = self.line_pos()
        tkn_value = "'"
        self.pop_char()
        while self.peek_char():
            tkn_value += self.peek_char()
            if self.peek_char() == "\n":
                self.pop_char()
                self.tokens.append(Token("TKN_ERROR", pos))
                return
            if self.peek_char() == "'":
                for i in tkn_value:
                    if i == '\\':
                        self.__line_pos += 1
                self.pop_char()
                self.tokens.append(Token("CHAR_CONST", pos, tkn_value))
                return
            self.pop_char()
        raise TokenError(pos)

    def constant(self):
        """Numeric constants can take many forms:
        - integer constants only allow digits [0-9]
        - real number constant only allow digits [0-9],
            ONE optionnal dot '.' and ONE optionnal 'e/E' character
        - binary constant only allow digits [0-1] prefixed by '0b' or '0B'
        - hex constant only allow digits [0-9], letters [a-f/A-F] prefixed
            by '0x' or '0X'
        - octal constants allow digits [0-9] prefixed by a zero '0'
            character

        Size ('l/L' for long) and sign ('u/U' for unsigned) specifiers can
        be appended to any of those. tokens

        Plus/minus operators ('+'/'-') can prefix any of those tokens

        a numeric constant could start with a '.' (dot character)
        """
        pos = self.line_pos()
        tkn_value = ""
        bucket = ".0123456789aAbBcCdDeEfFlLuUxX-+"
        while self.peek_char() and (self.peek_char() in bucket or self.peek_char() == "\\\n"):
            if self.peek_char() in "xX":
                if tkn_value.startswith("0") is False or len(tkn_value) > 1:
                    raise TokenError(pos)
                for c in "xX":
                    if c in tkn_value:
                        raise TokenError(pos)

            elif self.peek_char() in "bB":
                if tkn_value != "0" and tkn_value.startswith("0x") is False and tkn_value.startswith("0X") is False:
                    raise TokenError(pos)

            elif self.peek_char() in "+-":
                if (
                    tkn_value.endswith("e") is False
                    and tkn_value.endswith("E") is False
                    or self.peek_sub_string(2) in ["++", "--"]
                ):
                    break

            elif self.peek_char() in "eE" and "0x" not in tkn_value and "0X" not in tkn_value:
                if (
                    "e" in tkn_value
                    or "E" in tkn_value
                    or "f" in tkn_value
                    or "F" in tkn_value
                    or "u" in tkn_value
                    or "U" in tkn_value
                    or "l" in tkn_value
                    or "L" in tkn_value
                ):
                    raise TokenError(pos)

            elif self.peek_char() in "lL":
                lcount = tkn_value.count("l") + tkn_value.count("L")
                if (
                    lcount > 1
                    or (lcount == 1 and tkn_value[-1] not in "lL")
                    or ("f" in tkn_value or "F" in tkn_value)
                    and "0x" not in tkn_value
                    and "0X" not in tkn_value
                ):
                    raise TokenError(pos)
                elif self.peek_char() == "l" and "L" in tkn_value or self.peek_char() == "L" and "l" in tkn_value:
                    raise TokenError(pos)

            elif self.peek_char() in "uU":
                if (
                    "u" in tkn_value
                    or "U" in tkn_value
                    or (
                        ("e" in tkn_value or "E" in tkn_value or "f" in tkn_value or "F" in tkn_value)
                        and ("0x" not in tkn_value and "0X" not in tkn_value)
                    )
                ):
                    raise TokenError(pos)

            elif self.peek_char() in "Ff":
                if (
                    tkn_value.startswith("0x") is False
                    and tkn_value.startswith("0X") is False
                    and ("." not in tkn_value or "f" in tkn_value or "F" in tkn_value)
                    or "u" in tkn_value
                    or "U" in tkn_value
                    or "l" in tkn_value
                    or "L" in tkn_value
                ):
                    raise TokenError(pos)

            elif (
                self.peek_char() in "aAbBcCdDeE"
                and tkn_value.startswith("0x") is False
                and tkn_value.startswith("0X") is False
                or "u" in tkn_value
                or "U" in tkn_value
                or "l" in tkn_value
                or "L" in tkn_value
            ):
                raise TokenError(pos)

            elif (
                self.peek_char() in "0123456789"
                and "u" in tkn_value
                or "U" in tkn_value
                or "l" in tkn_value
                or "L" in tkn_value
            ):
                raise TokenError(pos)

            elif self.peek_char() == "." and "." in tkn_value:
                raise TokenError(pos)

            tkn_value += self.peek_char()
            self.pop_char()
        if tkn_value[-1] in "eE" and tkn_value.startswith("0x") is False or tkn_value[-1] in "xX":
            raise TokenError(pos)
        else:
            self.tokens.append(Token("CONSTANT", pos, tkn_value))

    def mult_comment(self):
        pos = self.line_pos()
        self.pop_char(), self.pop_char()
        tkn_value = "/*"
        while self.peek_char():
            if self.src[self.__pos :].startswith("*/"):
                tkn_value += "*/"
                self.pop_char(), self.pop_char()
                break
            tkn_value += self.peek_char()
            if self.peek_char() == "\n":
                self.__line += 1
                self.__line_pos = 1
            self.pop_char()
        if tkn_value.endswith("*/"):
            self.tokens.append(Token("MULT_COMMENT", pos, tkn_value))
        else:
            raise TokenError(pos)

    def comment(self):
        """Comments are anything after '//' characters, up until a newline or
        end of file
        """
        pos = self.line_pos()
        tkn_value = "//"
        self.pop_char(), self.pop_char()
        while self.peek_char() is not None:
            if self.peek_char() == "\n":
                self.tokens.append(Token("COMMENT", pos, tkn_value))
                return
            tkn_value += self.peek_char()
            self.pop_char()
        if (self.__pos == self.len):
            self.tokens.append(Token("COMMENT", pos, tkn_value))
            return
        raise TokenError(pos)

    def identifier(self):
        """Identifiers can start with any letter [a-z][A-Z] or an underscore
        and contain any letters [a-z][A-Z] digits [0-9] or underscores
        """
        pos = self.line_pos()
        tkn_value = ""
        while self.peek_char() and (
            self.peek_char() in string.ascii_letters + "0123456789_" or self.peek_char() == "\\\n"
        ):
            if self.peek_char() == "\\\n":
                self.pop_char()
                continue
            tkn_value += self.peek_char()
            self.pop_char()
        if tkn_value in keywords:
            self.tokens.append(Token(keywords[tkn_value], pos))

        else:
            self.tokens.append(Token("IDENTIFIER", pos, tkn_value))

    def operator(self):
        """Operators can be made of one or more sign, so the longest operators
        need to be looked up for first in order to avoid false positives
        eg: '>>' being understood as two 'MORE_THAN' operators instead of
            one 'RIGHT_SHIFT' operator
        """
        pos = self.line_pos()
        if self.peek_char() in ".+-*/%<>^&|!=":

            if self.peek_sub_string(3) in [">>=", "<<=", "..."]:
                self.tokens.append(Token(operators[self.peek_sub_string(3)], pos))
                self.pop_char(), self.pop_char(), self.pop_char()

            elif self.peek_sub_string(2) in [">>", "<<", "->"]:
                self.tokens.append(Token(operators[self.peek_sub_string(2)], pos))
                self.pop_char(), self.pop_char()

            elif self.peek_sub_string(2) == self.peek_char() + "=":
                self.tokens.append(Token(operators[self.peek_sub_string(2)], pos))
                self.pop_char(), self.pop_char()

            elif self.peek_char() in "+-<>=&|":
                if self.peek_sub_string(2) == self.peek_char() * 2:
                    self.tokens.append(Token(operators[self.peek_sub_string(2)], pos))
                    self.pop_char()
                    self.pop_char()

                else:
                    self.tokens.append(Token(operators[self.peek_char()], pos))
                    self.pop_char()

            else:
                self.tokens.append(Token(operators[self.peek_char()], pos))
                self.pop_char()

        else:
            self.tokens.append(Token(operators[self.src[self.__pos]], pos))
            self.pop_char()

    def preprocessor(self):
        pos = self.line_pos()
        tkn_value = ""
        while self.peek_char():
            tkn_value += self.peek_char()
            self.pop_char()
            if self.peek_sub_string(2) == "\\\n":
                self.__line_pos = 1
                self.__line += 1
            #    raise TokenError(self.line_pos())
            if self.peek_sub_string(2) in ["//", "/*"] or self.peek_char() == "\n":
                break
        if len(tkn_value) <= 1:
            raise TokenError(self.line_pos())
        tkn_key = tkn_value[1:].split()[0]
        if tkn_key not in preproc_keywords and tkn_key[: len("include")] != "include":
            raise TokenError(self.line_pos())
        else:
            if tkn_key not in preproc_keywords and tkn_key[: len("include")] == "include":
                tkn_key = "include"
            self.tokens.append(Token(preproc_keywords.get(tkn_key), pos, tkn_value))

    def get_next_token(self):
        """Peeks one character and tries to match it to a token type,
        if it doesn't match any of the token types, an error will be raised
        and current file's parsing will stop
        """
        while self.peek_char() is not None:
            if self.is_string():
                self.string()

            elif (self.peek_char().isalpha() and self.peek_char().isascii()) or self.peek_char() == "_":
                self.identifier()

            elif self.is_constant():
                self.constant()

            elif self.is_char_constant():
                self.char_constant()

            elif self.peek_char() == "#":
                self.preprocessor()

            elif self.src[self.__pos :].startswith("/*"):
                self.mult_comment()

            elif self.src[self.__pos :].startswith("//"):
                self.comment()

            elif self.peek_char() in "+-*/,<>^&|!=%;:.~?":
                self.operator()

            elif self.peek_char() == " ":
                self.tokens.append(Token("SPACE", self.line_pos()))
                self.pop_char()

            elif self.peek_char() == "\t":
                self.tokens.append(Token("TAB", self.line_pos()))
                self.pop_char()

            elif self.peek_char() == "\n":  # or ord(self.peek_char()) == 8203:
                self.tokens.append(Token("NEWLINE", self.line_pos()))
                self.pop_char()
                self.__line_pos = 1
                self.__line += 1

            elif self.peek_char() == "\\\n":
                self.tokens.append(Token("ESCAPED_NEWLINE", self.line_pos()))
                self.pop_char()
                self.__line_pos = 1
                self.__line += 1

            elif self.peek_char() in brackets:
                self.tokens.append(Token(brackets[self.peek_char()], self.line_pos()))
                self.pop_char()
            else:
                raise TokenError(self.line_pos())

            return self.peek_token()

        return None

    def get_tokens(self):
        """Iterate through self.get_next_token() to convert source code into a
        token list
        """
        while self.get_next_token():
            continue
        return self.tokens

    def print_tokens(self):
        if self.tokens == []:
            return
        for t in self.tokens:
            if t.type == "NEWLINE":
                print(t)
            else:
                print(t, end="")
        if self.tokens[-1].type != "NEWLINE":
            print("")

    def check_tokens(self):
        """
        Only used for testing
        """
        if self.tokens == []:
            self.get_tokens()
            if self.tokens == []:
                return ""
        ret = ""
        for i in range(0, len(self.tokens)):
            ret += self.tokens[i].test()
            ret += "" if self.tokens[i].type != "NEWLINE" else "\n"
        if self.tokens[-1].type != "NEWLINE":
            ret += "\n"
        return ret



from norminette.context import ControlStructure
from norminette.scope import Function
from norminette.context import GlobalScope
from norminette.scope import UserDefinedType
from norminette.rules import PrimaryRule
import pdb

lbrackets = ["LBRACE", "LPARENTHESIS", "LBRACKET"]
rbrackets = ["RBRACE", "RPARENTHESIS", "RBRACKET"]

misc_specifiers = [
    "CONST",
    "REGISTER",
    "STATIC",
    "VOLATILE",
    "EXTERN",
    "INLINE",
    "RESTRICT",
    "SIGNED",
    "UNSIGNED",
]

type_specifiers = [
    "CHAR",
    "DOUBLE",
    "ENUM",
    "FLOAT",
    "INT",
    "UNION",
    "VOID",
    "LONG",
    "SHORT",
    "STRUCT",
    "ENUM",
    "UNION",
]

class IsVarDeclaration(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 75
        self.scope = [GlobalScope, UserDefinedType, Function, ControlStructure]

    def assignment_right_side(self, context, pos):
        sep = ["COMMA", "SEMI_COLON", "ASSIGN"]
        i = context.skip_ws(pos, nl=True)
        while context.peek_token(i) and context.check_token(i, sep) is False:
            if context.check_token(i, lbrackets) is True:
                i = context.skip_nest(i)
            i += 1
        return True, i

    def var_declaration(self, context, pos, identifier=False):
        pclose = ["RPARENTHESIS", "NEWLINE", "SPACE", "TAB"]
        brackets = 0
        parenthesis = 0
        braces = 0
        i = pos
        ret_store = None
        ids = []
        while context.peek_token(i) is not None and context.check_token(i, ["SEMI_COLON"]) is False:
            if context.check_token(i, "IDENTIFIER") is True and braces == 0 and brackets == 0 and parenthesis == 0:
                identifier = True
                ids.append(context.peek_token(i))
            elif context.check_token(i, ["COMMENT", "MULT_COMMENT"]) is True:
                i += 1
                continue
            elif context.check_token(i, ["COLON", "CONSTANT"]) is True and identifier == True:
                i += 1
                continue
            elif context.check_token(i, lbrackets) is True:
                if context.check_token(i, "LBRACE") is True:
                    braces += 1
                if context.check_token(i, "LBRACKET") is True:
                    brackets += 1
                if context.check_token(i, "LPARENTHESIS") is True and brackets == 0 and braces == 0:
                    ret, tmp = context.parenthesis_contain(i, ret_store)
                    if ret == "function" or ret == "pointer" or ret == "var":
                        ret_store = ret
                        identifier = True
                        tmp2 = tmp - 1
                        deep = 1
                        while tmp2 > 0 and deep > 0:
                            if context.check_token(tmp2, "IDENTIFIER"):
                                ids.append(context.peek_token(tmp2))
                            if context.check_token(tmp2, "RPARENTHESIS"):
                                deep += 1
                            if context.check_token(tmp2, "LPARENTHESIS"):
                                deep -= 1
                            tmp2 -= 1
                        i = tmp
                    else:
                        parenthesis += 1
            elif context.check_token(i, rbrackets) is True:
                if context.check_token(i, "RBRACE") is True:
                    braces -= 1
                if context.check_token(i, "RBRACKET") is True:
                    brackets -= 1
                if context.check_token(i, "RPARENTHESIS") is True:
                    parenthesis -= 1
            elif context.check_token(i, "ASSIGN") is True:
                if identifier == False:
                    return False, pos
                ret, i = self.assignment_right_side(context, i + 1)
                i -= 1
                if ret is False:
                    return False, pos
            elif context.check_token(i, ["SPACE", "TAB", "MULT", "BWISE_AND", "NEWLINE"] + misc_specifiers + type_specifiers):
                pass
            elif context.check_token(i, "COMMA") is True and parenthesis == 0 and brackets == 0 and braces == 0:
                break
            elif parenthesis == 0 and brackets == 0 and braces == 0:
                return False, 0
            i += 1
        if identifier == False or braces > 0 or brackets > 0 or parenthesis > 0:
            return False, 0
        context.scope.vars_name.append(ids[-1])
        if context.check_token(i, "SEMI_COLON") is True:
            if brackets <= 0 and braces <= 0 and parenthesis <= 0:
                return True, i
            else:
                return False, 0
        if context.check_token(i, "COMMA") is True:
            i += 1
            return True, i
        return False, pos

    def is_func_pointer(self, context, pos):
        i = context.skip_ws(pos)
        ws = ["SPACE", "TAB", "NEWLINE"]
        if context.check_token(i, "LPARENTHESIS") is False:
            return False, pos
        identifier = False
        i += 1
        p = 1
        plvl = 0  # nesting level of the first pointer operator encountered

        while p and context.check_token(i, ["MULT", "LPARENTHESIS"] + ws):
            if context.check_token(i, "MULT") and not plvl:
                plvl = p
            elif context.check_token(i, "LPARENTHESIS"):
                p += 1
            i += 1

        while p and context.peek_token(i) is not None:
            if context.check_token(i, "LPARENTHESIS") is True:
                p += 1
                if identifier is True:
                    return False, pos
            elif context.check_token(i, "RPARENTHESIS") is True:
                p -= 1
                if identifier is True:
                    par_pos = i
                    break
            elif context.check_token(i, "IDENTIFIER") is True:
                identifier = True
            i += 1
        else:
            return False, pos
        i += 1
        i = context.skip_nest(i)
        return True, i

    def run(self, context):
        """
        Catches all kinds of variable declarations
        """
        ret, i = context.check_type_specifier(0)
        i = context.skip_ws(i)
        if ret is False:
            return False, 0
        tmp = i - 1
        while context.check_token(tmp, ["LPARENTHESIS", "MULT", "BWISE_AND"]):
            tmp -= 1
        if context.check_token(tmp, "SEMI_COLON"):
            return True, i
        if (
            context.check_token(tmp, ["SPACE", "TAB", "NEWLINE"]) is False
            and context.check_token(tmp - 1, ["SPACE", "TAB", "NEWLINE"]) is False
        ):
            return False, 0
        ret, i = self.var_declaration(context, i)
        if ret is False:
            return False, 0
        while ret:
            if context.check_token(i, "SEMI_COLON") is False:
                ret, i = self.var_declaration(context, i)
            if context.check_token(i, "SEMI_COLON") is True:
                i += 1
                i = context.eol(i)
                return True, i
        return False, 0
